package rwd.abstractsyntax.spiderdiagram;

import java.io.Serializable;
import java.util.Vector;
import java.util.Observable;

public abstract class Diagram 	
	extends Observable 
	implements Serializable
{
    private String name;

    public Diagram() { // just for reading from file
    }
    Diagram(String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }

    public void setName(String newName) {
        name = newName;
    }

    public void tellObservers() {
        setChanged();
        //System.out.println("notifying observers");
        this.notifyObservers();
    }

    public abstract boolean equals(Diagram d);

    public abstract String toString();
    /*
     * Looks for a part called the given name. Returns the first one found,
     * as the second component of an Vector. The first component is the parent
     * of this found part. It is sometimes useful to know the parent of a sought part
     * in rule applications. If the sought part is at the top of the tree,
     * the first component of the Vector is "null". If no such part is found,
     * a null Vector is returned.
     * To do : check this description.
     */
    public abstract Vector findPartCalled(String name);
    // returns parent,part

    /*
     * A flattened diagram has more than one component in each and-component
     * and or-component.  
     * In addition, each and-component has no children which are and-components.
     * In addition, each or-component has no children which are or-components.
     * This flattening massively reduces the number of reasoning manipulations required
     * to transform one diagram into another. However, it does make the code for inverting
     * some rules more complex, if flattening has occurred.
     */
    public abstract void flatten();

    /*
     * Clone diagrams, for example when a rule is applied - don't alter
     * the starting diagram but create a clone and edit that. 
     */
    public abstract Diagram getClone();

    //public abstract Vector findPartLike(Diagram d);// returns parent,part
    // was a daft idea - initial thinking was to look for "equal" but not
    // necessarily identical objects for tree parts - but a tree can have many
    // equal parts so this is a useless way of identifying parts from a clone, say.

    public abstract CompoundDiagram getParentOf(Diagram part);

    /*
     * Gets all names of all parts of the tree.
     */
    public abstract Vector getAllNames();

    /*
     * This method renames the parts of a compound diagram structure
     * ensuring that each part is identifiable by its name.  Names are unique
     * within the tree-structure and also different from the String objects provided
     * as parameter. (to do : find out how the names are adapted and document it here)
     */
    public abstract void renameOutside(Vector bannedNames);

    /*
     * Gets the "leaves" of the "tree" of a compound diagram. The returned
     * Vector contains UnitaryDiagram objects.  To do : check these leaves
     * are identically equal to the leaves aand not clones.
     */
    public abstract Vector getUnitaryParts();

    /*
     * Gets the "parts" of the "tree" of a compound diagram. The returned
     * Vector contains Diagram objects. For example, a diagram of the form
     * P and (Q or R) would have four parts : the whole thing, the unitary part, 
     * leaf, P, the compound part Q or R, and two more leaves, UnitaryDiagrams
     * Q and R.  
     * To do : check these leaves
     * are identically equal to the leaves aand not clones.
     */
    public abstract Vector getParts();

    protected abstract void addUnitaryComponents(Vector componentsSoFar);

    /*
     * Coordinates are used to identify parts in compound diagram structures.
     * The int array {} represents the top diagram = the whole diagram.
     * The int array {2} represents the 2nd child of the whole diagram - in order for this
     * to make sense, the whole diagram must be an n-ary diagram.
     * To do - check what this method returns if the part isn't part of the whole
     * anyway.
     */
    public abstract int[] getCoords(Diagram part);

    /*
     * Coordinates are used to identify parts in compound diagram structures.
     * The int array {} represents the top diagram = the whole diagram.
     * The int array {2} represents the 2nd child of the whole diagram - in order for this
     * to make sense, the whole diagram must be an n-ary diagram.
     * To do - check what this method returns if the coords are inappropriate for
     * the diagram structure.
     */
    public abstract Diagram getPartAt(int[] coords);

    /*
     * In general, deduction rules can only be applied to parts
     * which are not nagetd in a whole expression.  So, in P or Not(Q),
     * we should be able to remove shading from P but not from Q.
     * This method takes a part and determines whether or not it's
     * negated (an odd number of times) in the whole.
     * Three possible outcomes:
     * 1. The part is present, with an odd number of negations between
     * the part and the whole : return true
     * 2. The part is present, with an even number of negations between
     * the part and the whole : return false
     * 3. The part is absent : return false (should be useless to anyone anyway)
     */
    public abstract boolean containsNegationOf(Diagram part);
}
