package rwd.abstractsyntax.spiderdiagram;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;
import java.util.List;

import rwd.abstractsyntax.eulerdiagram.*;

/**
 * An abstract spider.
 * An abstract spider has a habitat, which is a set of zones, and
 * a label (which may be of zero length).
 * Different spiders in a unitary diagram should be distinguished by 
 * different labels.
 * Although abstract spiders have an associated Spider Library,
 * this is purely to save space and time when duplicating spiders;
 * an abstract spider is not a flyweight, and it is possible to
 * clone abstract spiders (i.e. create two separate spider objects
 * which are equal). 
 * @author rwd, Richard Bosworth
 * 
 * @version 8.7.05   08:56
 */
public class Spider
	implements Serializable
{
	/**
	 * Constructor for anonymous spider.
	 * Post: spider has been constructed with specified habitat
	 *       and a zero-length label.
	 * @param habitat = habitat of this spider as a vector of zones
	 */
	/*
	 * test: once thru
	 */
	public Spider(Vector habitat) {
		this("",habitat);
	}

	/**
	 * Constructor for spider with label.
	 * Post: spider has been constructed with specified habitat
	 *       and specified label.
	 * @param habitat habitat of this spider as a vector of zones
	 * @param label label of spider as string (may be zero-length)
	 */
	/*
	 * test: once thru
	 */
	public Spider(String label, Vector habitat) {
		setHabitat(habitat);
		setLabel(label);
	}

	public Vector getHabitat() {
		return habitat;
	}
	public void setHabitat(Vector habitat) {
		this.habitat = habitat;
	}

	public void addFoot(Zone z) {
		habitat.add(z);
	}

	/**
	 * Post: returned spider is in the Spider Library.
	 * @param label label of spider (may be zero-length)
	 * @param habitat list of zones representing this spider's habitat
	 * @return abstract spider with given specification
	 */
	public static Spider getAbstractSpider(String label, List habitat) {
		// scan through the library comparing spiders there with requirements
		Spider newSpider = new Spider(new Vector(habitat));
		newSpider.setLabel(label);
		for (int i = 0; i < spiderLibrary.size(); i++) {
			Spider spider = (Spider) (spiderLibrary.get(i));
			if (spider.equals(newSpider)) {
				return spider;
			}
		}
		// Spider wasn't in library so add it to the library
		spiderLibrary.add(newSpider);
		return newSpider;	
	}

	/**
	 * 
	 * @return string representing label of this spider 
	 *         (may be zero-length)
	 */
	/*
	 * test: once thru
	 */
	public String getLabel() {
		return label;
	}
	
	/**
	 * Post: label of this spider = s
	 * @param s label of this spider (may be zero-length)
	 */
	
	public void setLabel(String s) {
		label = s;
	}

	public boolean isUniversal() {
		return kind == UNIVERSAL;
	}	
	public boolean isExistential() {
		return kind == EXISTENTIAL;
	}	

	public boolean inDomain(UnitaryDiagram cud, Contour c) {
		List habitat = getHabitat();
		Iterator i = habitat.iterator();
		while (i.hasNext()) {
			Zone az = (Zone)i.next();
			if (az.getContainingContours().contains(c)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 
	 * @param o object to be compared with this spider
	 * @return true iff o is a spider and this spider and o have the same 
	 *         habitat and the same string for a label
	 */
	/*
	 * test:
	 *   o is not a spider
	 *   o and this spider have a different habitat
	 *   o and this spider have the same habitat
	 *     o and this spider have different labels
	 *     o and this spider have the same label
	 */
	/**
	 * @deprecated 
	 */
	public boolean equals(Object o) {
		if (!(o instanceof Spider)) return false;
		Spider s = (Spider) o;
		return habitat.containsAll(s.habitat) && 
		       s.habitat.containsAll(habitat) && 
			   label.equals(s.getLabel());
	}

	/**
	 * 
	 * @return hashCode for this spider.  It is guaranteed that two equal spiders 
	 *         will have the same hashCode.
	 */
	/*
	 * test:
	 *   spider with empty habitat
	 *   spider with habitat of two zones
	 */
	public int hashCode() {
		int h = this.getLabel().hashCode();
		Collection habitat = this.getHabitat();
		Iterator hit = habitat.iterator();
		while (hit.hasNext())
			h += hit.next().hashCode();
		return h;
	}

	/**
	 * @return string representation of this spider
	 */
	/*
	 * test: once thru
	 */
	public String toString() {
		return "("+label+","+habitat+")";
	}
	
	/*
	 * pre:  sp = spider object to be searched for in library
	 * post: returns true iff sp is identically present in library
	 * 
	 */
	/*
	 * test:
	 *   sp is in library
	 *   sp is not in library
	 */
	public static boolean isInSpiderLibrary(Spider sp){
		Iterator split = spiderLibrary.iterator();
		while (split.hasNext()){
			Spider splsp = (Spider) split.next();
			if (splsp==sp) return true;
		}
		return false;		
	}
	

	// Local properties
	// ----------------
	
	// Set of abstract zones
	protected Vector habitat = new Vector();
	// Label
	protected String label;
	// Kind
	public int kind = EXISTENTIAL;
	// Kind
	public static int EXISTENTIAL = 0;
	public static int UNIVERSAL = 1;
	
	// Set of abstract spiders
	protected static Vector spiderLibrary = new Vector();

	
    // test program
    // ------------
    /*
	public static void main(String args[]) {
		Vector h = new Vector();
		h.add("test");
		Spider s = new Spider(h);
		Vector h2 = s.getHabitat();
		h2.remove("test");
	}
	*/
	
	/*
    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }
    
    public static void main(String args[]) throws Exception {
    	/*
    	{
			hd("Tests of isInSpiderLibrary");
			
			Vector va = new Vector();
			va.add(Contour.getContour("a"));
			pr("va="+va);
			
			Vector vb = new Vector();
			vb.add(Contour.getContour("b"));
			pr("vb="+vb);
			
			Zone zab = Zone.getZone(va,vb);
			pr("zab=" + zab + "  zab.hashCode()="+zab.hashCode());
			Zone zba = Zone.getZone(vb,va);
			pr("zba=" + zba + "  zba.hashCode()="+zba.hashCode());
			
			Vector hab1 = new Vector();
			hab1.add(zab);
			hab1.add(zba);
			
			pr("make a legitimate spider");
			Spider sp1 = Spider.getAbstractSpider("sp",hab1);
			pr("sp1="+sp1+",sp1.hashCode()="+sp1.hashCode());
			pg();
			
			pr("make a rogue spider");
			Spider sp2 = new Spider(hab1);
			sp2.setLabel("sp");
			pr("sp2="+sp2+",sp2.hashCode()="+sp2.hashCode());
			pg();
			  
			pr("Spider.isInSpiderLibrary(sp1)="+
				    Spider.isInSpiderLibrary(sp1));
				pg();
				  
			pr("Spider.isInSpiderLibrary(sp2)="+
			    Spider.isInSpiderLibrary(sp2));
			pg();
		}
		
    	{
    		hd("Test of constructors");
    		Contour a = Contour.getContour("a");
    		Vector v = new Vector();
    		Vector va = new Vector();
    		va.add(a);
    		pr("v="+v);
    		pr("va="+va);
    		Zone za = Zone.getZone(v,va);
    		pr("za="+za);
    		pg();
    		
    		Vector h = new Vector();
    		h.add(za);
    		pr("h="+h);
    		Spider s1 = new Spider(h);
    		pr("s1="+s1);
    		pg();
    		
    		Spider s2 = new Spider("fred",h);
    		pr("s2="+s2);
    		pg();
    		
    		
    	}
    	*/
    	/*
    	{
    		hd("Tests of equals");
    		Vector h = new Vector();
    		Object z1 = new Object();  // doesn't need to be a real zone
    		h.add(z1);
    		Spider s1 = new Spider("fred",h);
    		pr("s1="+s1);
    		pg();
    		
    		Object o1 = new Object();
    		pr("o1="+o1);
    		pr("s1.equals(o1)="+s1.equals(o1));
    		pg();
    		
    		Vector h2 = new Vector();
    		Spider s2 = new Spider("fred",h2);
    		pr("s2="+s2);
    		pr("s1.equals(s2)="+s1.equals(s2));
    		pg();
    		
    		Spider s3 = new Spider("jim",h);
    		pr("s3="+s3);
    		pr("s1.equals(s3)="+s1.equals(s3));
    		pg();
    		
    		Spider s4 = new Spider("fred",h);
    		pr("s4="+s4);
    		pr("s1.equals(s4)="+s1.equals(s4));
    		pg();
    		
    		
    	}
    	*/
		/*
    	{
    		hd("Tests of hashCode");
    		Vector eh = new Vector();
    		Vector h2 = new Vector();
    		Object z1 = new Object(); // no need to be real zones
    		Object z2 = new Object();
    		h2.add(z1);
    		h2.add(z2);
    		
    		Spider s1 = new Spider("bert",eh);
    		pr("s1="+s1);
    		pr("s1.hashCode()="+s1.hashCode());
    		Spider s2 = new Spider("bert",eh);
    		pr("s2="+s2);
    		pr("s2.hashCode()="+s2.hashCode());
    		pr("s1.equals(s2)="+s1.equals(s2));
    		pg();
    		
    		Spider s3 = new Spider("letitia",h2);
    		pr("s3="+s3);
    		pr("s3.hashCode()="+s3.hashCode());
    		Spider s4 = new Spider("letitia",h2);
    		pr("s4="+s4);
    		pr("s4.hashCode()="+s4.hashCode());
    		pr("s3.equals(s4)="+s3.equals(s4));
    		pg();
    	}
    	*/
	/*
    }
    */

}
