package rwd.concretesyntax;

import java.awt.*;
import java.awt.geom.*;
import java.util.HashSet;
import java.util.Set;

import rwd.concretesyntax.eulerdiagram.ConcreteContour;
import rwd.gui.editor.DiagramCanvas;

/**
 * An item in a concrete diagram.
 * <P>
 * Every item in a diagram has the following attributes:
 * <UL>
 *   <LI>a shape
 *   <LI>a position (center of shape)
 *   <LI>a line thickness
 *   <LI>a foreground color
 *   <LI>a background color
 *   <LI>a label, with a position, font and color (may be zero-length)
 *   <LI>the property of being attached to another item
 *   <LI>an id, which is a user-friendly description of the item
 * </UL>
 * <P>
 * Every item has the following operations:
 * <UL>
 *   <LI>autoPlaceLabel, which attempts to place the label of the item
 *   <LI>getClone, which returns a DiagramItem and does not throw an exception
 * </UL>
 * <P>
 * Note that a clone is <I>not</I> equal to its original.  Equality is defined 
 * as identity for concrete diagram items.
 * 
 * Class invariants:
 *   x.equals(y) implies x.hashCode() == y.hashCode()
 * 
 * @author  op,rnb
 * @version 11.8.05   13:13
 */
public class DiagramItem 
	extends Observable
	implements java.io.Serializable
{
    
    // Constructor
    // -----------
	
    public DiagramItem() {
    }

    
    // attributes
    // ----------
    
	public Color getBackColor() {
		return _backColor;
	}

    /**
     * Note: unlike clone(), this returns a DiagramItem, and does not throw
     * an exception.
     * @return a deep clone of this diagram item
     */
    /*
     * test: once thru
     */
	public DiagramItem getClone() {
		DiagramItem di = new DiagramItem();
		di.setAttributes(this);
		return di;
	}
	
	public Font getFont() {
		return _font;
	}
	
	public Color getForeColor() {
		return _foreColor;
	}

	public Set getHandles() {
    	Set s = new HashSet();
    	return s;
    }

    public String getLabel() {
        return _label;
    }
    
	/**
	 * @return position of label (middle of base)
	 */
	/*
	 * test: once thru with non-zero values
	 */
    public Point2D getLabelPosition() {
        return new Point2D.Float(_labelX, _labelY);
    }  
    
    public Shape getShape() {
    	return null;
    }

	public Shape getShape(boolean smooth) {
		return null;
	}
    
	public Color getTextColor() {
		return _textColor;
	}

    public float getThickness() {
        return _thickness;
    }

	/**
	 * 
	 * @return a user-friendly description of this item
	 */
	public String id() {
		return "diagram item " + getLabel();
	}
	
	public boolean isAttached() {
		return isAttached;
	}

	
    // operations
    // ----------
    
    /**
     * post: label has been placed in appropriate position for this item
     *
     */
	public void autoPlaceLabel() {
		_labelX = _centerX;
		_labelY = _centerY;
	}

	/**
	 * post: label of this item has been drawn on g
	 * @param g graphics object for drawing this item
	 */
	public void drawLabel(Graphics g) {
		// Do not write label for derived contours
		if(this instanceof ConcreteContour){
			ConcreteContour cp = (ConcreteContour)this;
			if (cp.isDerived()) return;
		}
		// Test if label is empty
		String text = getLabel();
		if (text == null || text.length() == 0)
			return;
		
		// Get label position
		Point2D point = getLabelPosition();
		int msgX = (int)point.getX();
		int msgY = (int)point.getY();
		
		// Draw text
		if (_font == null) _font = g.getFont();
		Font backup = g.getFont();
		
		// PJR changed to allow fading
		g.setColor(_foreColor);
		g.setFont(_font);
		String res = replaceUnicodeSequences(text);
		
		// calculate start of string position
		int stringWidth = g.getFontMetrics().stringWidth(res);
		msgX -= stringWidth/2;
		
		// draw string on screen
		g.drawString(res, msgX, msgY);
		g.setFont(backup);
	}

	public void setAttached(boolean b) {
		isAttached = b;
	}

    /**
     * post: this item's attributes are set to those of newItem
     * @param newItem diagram item whose attributes are to be copied
     */
	/*
	 * test: once thru with non-standard attributes
	 */
    public void setAttributes(DiagramItem newItem) {
    	this._label = new String(newItem._label);
    	this.setLabelPosition(newItem.getLabelPosition());
        this._backColor = newItem._backColor;
        this._foreColor = newItem._foreColor;
        this._font = newItem._font;
        this._thickness = newItem._thickness;
        this._textColor = newItem._textColor;               
    }

public void setForeColor(Color foreColor) {
        _foreColor = foreColor;
    }
    
	public void mixBackColor(Color color) {
		_backColor = color;		
//		if (_backColor == null) _backColor = Color.WHITE;
//		int r1 = _backColor.getRed();
//		int g1 = _backColor.getGreen();
//		int b1 = _backColor.getBlue();
//		int r2 = color.getRed();
//		int g2 = color.getGreen();
//		int b2 = color.getBlue();
//		int rgb1 = _backColor.getRGB();
//		int rgb2 = color.getRGB();
//		_backColor = new Color(r1&r2, g1&g2, b1&b2);
////		_backColor = new Color(rgb1 & rgb2);
	}
	
    public void setBackColor(Color backColor) {
        _backColor = backColor;
    }
    
	public void setFont(Font font) {
		this._font = font;
	}

    /**
     * Post: label of this item has been set to required value.
     *       If old value was empty, new label has been autopositioned.
     * @param label required value of label (may be empty)
     */
    /*
     * test:
     *   old label empty
     *   old label non-empty
     */
    public void setLabel(String label) {
    	if (_label.equals(""))
    		autoPlaceLabel();
        _label = label;
    }
    
    /**
     * Post: if point was non-null, label has been set to required position
     * @param point required position of centre of label's base
     */
    /*
     * test:
     *   point null
     *   point non-null
     */
    public void setLabelPosition(Point2D point) {
    	if (point != null) {
			float x = (float) point.getX();
			float y = (float) point.getY();
			_labelX = x;
			_labelY = y;
    	} else {
    		autoPlaceLabel();
    	}
    }
    
    /*
     * test: once thru
     */
	public void setLocation(float x, float y) {
		translateTo(x-_centerX,y-_centerY);
	}
	    
	public void setTextColor(Color color) {
		_textColor = color;
	}
        
    public void setThickness(float thickness) {
        _thickness = thickness;
    }

    /*
     * post: position of this item (including any label) has been translated by dx,dy
     * @param dx amount position is to be translated in x direction
     * @param dy amount position is to be translated in y direction
     */
    /*
     * test: once thru with non-zero dx,dy
     */
	public void translateTo(float dx, float dy) {
		_centerX += dx;
		_centerY += dy;
		_labelX += dx;
		_labelY += dy;
	}

	protected void translateLabelTo(float dx, float dy) {
		_labelX += dx;
		_labelY += dy;
	}

	// static methods
	// --------------
	
	public static float median(int x[], int n) {
		float s = 0;
		for(int i=0; i<n; i++) {
			s += x[i];
		}
		s /= n;
		return s;
	}
	
	
	public static String replaceUnicodeSequences(String s) {
		String res = "";
		boolean inSequence = false;
		if (s == null) return res;
		int n = s.length();
		for(int i=0; i<n; i++) {
			char ch = s.charAt(i);
			if (ch == '\\') {
				char value = 0;
				boolean ok = true;
				if (i<n && Character.toLowerCase(s.charAt(i+1)) == 'u') {
					for(int j=0; j<4; j++) {
						if (i+2+j < n) {
							char ch1 = s.charAt(i+2+j);
							int digit = Character.digit(ch1, 16);
							value = (char)(value*16 + digit);
						} else {
							ok = false;
						}
					} 
				} else {
					ok = false;
				}
				if (ok) {
					res += value;
					i += 5;
				}
			} else {
				res += ch;
			}
		}
		return res;
	}
	
	// local methods
	// -------------
	
	/*
	 * pre:  p = polygon whose string is to be returned
	 * post: returns string of p, giving coordinates of points
	 * 
	 * Note: this method is useful because the toString method for Polygon
	 *       does not display the points of the polygon.
	 */
	/*
	 * test:
	 *   polygon with no points
	 *   polygon with two points
	 */
	
	protected static String stringOf(Polygon p) {
		String s = "Polygon[";
		for (int i=0; i < p.npoints; i++) {
			s += "("+ p.xpoints[i] + "," + p.ypoints[i] + ")";
			if (i<p.npoints-1) s += ",";
		}
		s += "]";		
		return s;		
	}


	// Local properties
	// ----------------
	
    protected Color _foreColor = Color.black;
	protected Color _backColor = Color.white;
	protected Font _font = rwd.gui.editor.EditorFrame.options.getFont();
	protected float _thickness = DiagramCanvas.DEFAULT.getLineWidth();
	protected Color _textColor = Color.black;
	protected boolean isAttached;
	
    // Label position
	protected String _label = "";
    protected float _labelX = 0;  // x-position of centre of base of label
    protected float _labelY = 0;  // y-position of centre of base of label
    
	// Center position
	protected float _centerX = 0;
	protected float _centerY = 0;
	
	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	
    	DiagramItem di = new DiagramItem();
    	di.setLabel("di");
    	pr("di="+di);
    	pr("di.id()="+di.id());
    	pg();
    	
    	/*
    	{
    		hd("Test of equals");
        	pr("di.equals(null)="+di.equals(null));
        	pg();
    		
    		Object o = new Object();
    		pr("o="+o);
        	pr("di.equals(o)="+di.equals(o));
        	pg();
    		
    		DiagramItem di2 = di.getClone();
        	pr("di2="+di2);
        	pr("di.equals(di2)="+di.equals(di2));
        	pg();
        	
        	pr("di.equals(di)="+di.equals(di));
        	pg();
    	}
    	
    	{
    		hd("Test of stringOf");
        	Polygon p = new Polygon();
        	pr("p="+stringOf(p));
        	pg();

        	Polygon p1 = new Polygon();
        	p1.addPoint(1,2);
        	p1.addPoint(3,4);
        	pr("p1="+stringOf(p1));
        	pg();
    	}
    	
    	{
    		hd("Test label positioning");
    		pr("di="+di);
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.setLocation(1,2)");
    		di.setLocation(1,2);
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.setLabelPosition(new Point(3,4))");
    		di.setLabelPosition(new Point(3,4));
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.autoPlaceLabel()");
    		di.autoPlaceLabel();
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.setLabelPosition(new Point(5,6))");
    		di.setLabelPosition(new Point(5,6));
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.setLabel(\"\")");
    		di.setLabel("");
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.setLabel(\"fred\")");
    		di.setLabel("fred");
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.translateTo(1,2)");
    		di.translateTo(1,2);
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    		pr("di.translateLabelTo(1,2)");
    		di.translateLabelTo(1,2);
        	pr("di.id()="+di.id());
    		pr("di._centerX="+di._centerX);
    		pr("di._centerY="+di._centerY);
    		pr("di._labelX="+di._labelX);
    		pr("di._labelY="+di._labelY);
    		pr("di.getLabelPosition()="+di.getLabelPosition());
    		pg();
    		
    	}
    	
    }
	*/
	
}
