package rwd.gui.editor.listeners;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.Iterator;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import rwd.gui.editor.EditorFileFilter;
import rwd.gui.editor.EditorFrame;
import rwd.gui.editor.EditorInternalFrame;
import rwd.gui.editor.viewers.DiagramViewerInterface;
import rwd.gui.editor.writers.DiagramWriter;
import rwd.gui.editor.writers.EpsWriter;
import rwd.gui.editor.writers.ImageWriter;
import rwd.gui.editor.writers.SerialWriter;
import rwd.gui.editor.writers.XmlWriter;
import uk.ac.kent.cs.kmf.util.ILog;

/**
 * @author Octavian Patrascoiu
 *
 */
public class FileSaveListener implements ActionListener {
	public FileSaveListener(EditorFrame frame) {
		this.frame = frame;
		this.saveAs = false;
		this.extension = "cd";
		this.log = frame.getLog();
	}

	public FileSaveListener(EditorFrame frame, boolean saveAs) {
		this.frame = frame;
		this.saveAs = saveAs;
		this.extension = "cd";
		this.log = frame.getLog();
	}

	public FileSaveListener(EditorFrame frame, boolean saveAs, String extension) {
		this.frame = frame;
		this.saveAs = saveAs;
		this.extension = extension;
		this.log = frame.getLog();
	}

	public void actionPerformed(ActionEvent e) {
		// Get current internal frame and viewer
		EditorInternalFrame editor = frame.getCurrentEditor();
		DiagramViewerInterface viewer = frame.getCurrentViewer();
		if (frame != null && viewer != null) {
			File file = editor.getFile();
			// Save
			if (!saveAs) {
				// If not saved before choose file
				if (file == null) {
					// Create chooser
					JFileChooser chooser = new JFileChooser(frame.getDirectory());
					chooser.setAcceptAllFileFilterUsed(false);
					chooser.setDialogTitle("Save diagram");
					chooser.setApproveButtonText("Save");
					chooser.setApproveButtonToolTipText("Save file");
					// Add filters
					EditorFileFilter filter = addFilter(chooser, "Serialized Format (.cd)", "cd");
					addFilter(chooser, "Extended Markup Language (.xml)", "xml"); 
					chooser.setFileFilter(filter);
					int returnVal = chooser.showOpenDialog(null);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						file = chooser.getSelectedFile();
						// Save it
						String extension = ((EditorFileFilter) chooser.getFileFilter()).getFirstExtension();
						file = save(extension, file, viewer);
						editor.setFile(file);
						// Set title
						editor.setTitle(file.getName());
					} else {
						return;
					}
				// Already saved
				} else {
					// Save it
					String fileName = file.getAbsolutePath();
					int pos = fileName.lastIndexOf('.');
					String extension = "cd";
					if (pos != -1) extension = fileName.substring(pos+1, fileName.length());
					file = save(extension, file, viewer);
					editor.setFile(file);
				}
			// Save as
			} else {
				// Create file chooser
				JFileChooser chooser = new JFileChooser(frame.getDirectory());
				chooser.setAcceptAllFileFilterUsed(false);
				chooser.setDialogTitle("Save As");
				chooser.setApproveButtonText("Save As");
				chooser.setApproveButtonToolTipText("Save file");
				// Add filters
				EditorFileFilter filter = addFilter(chooser, "Serialized Format (.cd)", "cd");
				addFilter(chooser, "Extended Markup Language (.xml)", "xml"); 
				addFilter(chooser, "Extended Postscript Format (.eps)", "eps");
				addFilter(chooser, "Graphics Interchange Format (.gif)", "gif");
				addFilter(chooser, "Windows Bitmap (.bmp)", "bmp");
				addFilter(chooser, "Joint Photographic Experts Group (.jpg)", "jpg");
				addFilter(chooser, "Portable Network graphics (.png)", "png");
				chooser.setFileFilter(filter);
				// Choose it
				int returnVal = chooser.showOpenDialog(null);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					file = chooser.getSelectedFile();
					String extension = ((EditorFileFilter) chooser.getFileFilter()).getFirstExtension();
					// Save it
					save(extension, file, viewer);
					// Set title
					editor.setTitle(file.getName());
				}
			}
		}
	}
	
	protected EditorFileFilter addFilter(JFileChooser chooser, String name, String ext) {
		EditorFileFilter filter = new EditorFileFilter(name);
		filter.addExtension(ext);
		chooser.addChoosableFileFilter(filter);
		return filter;
	}
	
	protected File save(String ext, File file, DiagramViewerInterface viewer) {
		// Prepare file name
		String filename = file.getName();
		if (!filename.toLowerCase().endsWith("."+ext)) {
			filename = filename + "."+ext;
			file = new File(file.getParentFile(), filename);
		}
		// Update editor's current directory
		frame.setDirectory(file.getParentFile());
		// Save it according to extension
		DiagramWriter writer = null;
		if (ext.equals("xml"))
			writer = new XmlWriter(log);
		else if (ext.equals("cd"))
			writer = new SerialWriter(log);
		else if (ext.equals("eps"))
			writer = new EpsWriter(log);
		else 
			writer = new ImageWriter(log, ext);
		writer.write(file, viewer);
		return file;
	}
	

	public File saveEPS(File file, DiagramViewerInterface viewer) {
		// Save the file
		try {
			BufferedWriter output =
				new BufferedWriter(new FileWriter(file));
			output.write(viewer.getDiagramCanvas().getEPS());
			output.flush();
			output.close();
		} catch (Exception anye) {
			JOptionPane.showMessageDialog(
				null,
				"The EPS file could not be created:\n\n" + anye,
				"Could not write EPS",
				JOptionPane.ERROR_MESSAGE);
			return null;
		}
		return file;
	}

	public File saveXML(File file, DiagramViewerInterface viewer) {
		// Save the file
		try {
			// Create an XML writer and write the diagrams
			DiagramWriter writer = new XmlWriter(log);
			if (viewer != null)
				writer.write(file.getAbsolutePath(), viewer);
		} catch (Exception anye) {
			JOptionPane.showMessageDialog(
				null,
				"The XML file could not be created:\n\n" + anye,
				"Could not write XML",
				JOptionPane.ERROR_MESSAGE);
			return null;
		}
		return file;
	}

	public File saveImage(String suffix, File file, DiagramViewerInterface viewer) {
		// Check if the platform supports the suffix
		Iterator imageWriterIt =
			ImageIO.getImageWritersBySuffix(suffix);
		if (!imageWriterIt.hasNext()) {
			JOptionPane.showMessageDialog(
				null,
				"Your Java platform does not support the writing of '"
					+ suffix
					+ "' files.",
				"Could not write image",
				JOptionPane.ERROR_MESSAGE);
			return null;
		}
		// Save the file
		BufferedImage image =
			viewer.getDiagramCanvas().getBufferedImage(viewer.getBoxAround(0));
		try {
			ImageIO.write(image, suffix, file);
		} catch (Exception anye) {
			JOptionPane.showMessageDialog(
				null,
				"The image could not be written:\n\n" + anye,
				"Could not write image",
				JOptionPane.ERROR_MESSAGE);
			return null;
		}
		return file;
	}

	// Local variables
	protected boolean saveAs;
	protected String extension;
	protected EditorFrame frame;
	protected ILog log;
}
