package rwd.gui.editor.listeners;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;
import java.util.Vector;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.concretesyntax.DiagramItem;
import rwd.gui.editor.EditorFrame;
import rwd.metrics.*;
import rwd.metrics.quality.QualityModel;
import rwd.metrics.quality.viewers.Viewer;
import uk.ac.kent.cs.kmf.util.ILog;

/**
 * @author Octavian Patrascoiu
 *
 */
public class MetricsComputeListener implements ActionListener {
	public MetricsComputeListener(EditorFrame frame) {
		this.frame = frame;
	}

	public void actionPerformed(ActionEvent e) {
		// Create a dialog to choose diagram and metrics
		MetricsDialog dialog =
			new MetricsDialog(frame, "Metrics", frame.getLog());
		int returnVal = dialog.showOpenDialog(frame);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			String inputString = dialog.getDiagramName();
			if (inputString != null) {
				// Get the selected concrete diagram
				ConcreteUnitaryDiagram cud = frame.getCUDCalled(inputString);
				
				if (cud == null) {
					JOptionPane.showMessageDialog(
						frame,
						"can't identify unitary diagram called "
							+ inputString);
					return;
				}


				// Get the metric names
				List metricNames = dialog.getMetricsName();

				// Make matrics
				List metrics = makeMetrics(cud, metricNames);

				// Activate deltas to compute metrics
				activateDeltas(metrics);

				// Create a metrics evaluator
				MetricsEvaluator eval = new MetricsEvaluator(cud, metrics);

				// Generate the result of the evaluation
				String res = eval.evaluate();

				// Deactivate deltas
				if (!dialog.isActiveDeltas()) {
					deactivateDeltas(cud);
				}

				// Create html and launch browser
				viewReport(cud, metrics, frame.getLog());
				// Reset log
				frame.getLog().reset();

				// Print it out
				frame.getLog().reportMessage(res);
			}
		}
	}
	
	protected void viewReport(ConcreteUnitaryDiagram cud, List metrics, ILog log) {
		try {
			// Get selected elements
			QualityModel qModel = new QualityModel(cud, metrics, log);
			Viewer viewer = new Viewer(qModel, cud, Viewer.KIVIAT);
			viewer.view();
		} catch (Exception ex) {
			ex.printStackTrace();
			log.reportMessage(""+ex);
		}
	}


	protected List makeMetrics(ConcreteUnitaryDiagram cud, List names) {
		List res = new Vector();
		if (names.contains("NodeNodeSeparationMetric"))
			res.add(new NodeNodeSeparation(cud, 1.0));
		if (names.contains("NodeContourSeparationMetric"))
			res.add(new NodeContourPointSeparation(cud, 1.0));
		if (names.contains("EdgeLengthMetric"))
			res.add(new EdgeLength(cud, 1.0));
		if (names.contains("AngularResolutionMetric"))
			res.add(new AngularResolution(cud, 1.0));
		if (names.contains("EdgeEdgeCrossingMetric"))
			res.add(new EdgeEdgeCrossing(cud, 1.0));
		if (names.contains("EdgeContourCrossingMetric"))
			res.add(new EdgeContourCrossing(cud, 1.0));
		return res;
	}

	protected void activateDeltas(List metrics) {
		// For each metric
		for (int j = 0; j < metrics.size(); j++) {
			Metric m = (Metric) metrics.get(j);
			m.activateDeltaObservers();
		}

		// Print it out
		frame.getLog().reportMessage("Delta observers are activated !");
	}

	protected void deactivateDeltas(ConcreteUnitaryDiagram cud) {
		// For each diagram item
		List items = cud.getDiagramItems();
		for (int i = 0; i < items.size(); i++) {
			// Remove observers
			DiagramItem item = (DiagramItem) items.get(i);
			item.deleteObservers();
		}

		// Print it out
		frame.getLog().reportMessage("Delta observers are deactivated !");
	}

	// Local variables
	protected EditorFrame frame;
}
