package rwd.ocl.constraintdiagram;

import java.util.*;

/**
 * An abstract representation of a reading tree node.
 * A node has an optional father and zero or more sons.
 * 
 * @author Octavian Patrascoiu, rnb
 * @version 5.9.05   13:18
 *
 */
public class ReadingTreeNode {
	
	// Constructors
	// ------------
	
	/**
	 * Post: creates reading tree node with no sons
	 * @param info name of this node
	 */
	public ReadingTreeNode(String info) {
		this.info = info;
		this.sons = new ArrayList();
	}
	
	/**
	 * Post: creates reading tree node with specified sons
	 * @param info name of this node
	 * @param sons sons of this node (must have this node as father)
	 */
	public ReadingTreeNode(String info, List sons) {
		this.info = info;
		this.sons = sons;
	}
	
	// attributes
	// ----------
	
	/**
	 * @return info string of this reading tree node
	 */
	/*
	 * test: once thru with non-empty info
	 */
	public String getInfo() {
		return info;
	}
	
	/**
	 * @return father of this reading tree node
	 */
	/*
	 * test: once thru with non-null father
	 */
	public ReadingTreeNode getFather() {
		return father;
	}
	
	/**
	 * @return list of sons of this reading tree node
	 */
	/*
	 * test: once thru with one or more sons
	 */
	public List getSons() {
		return sons;
	}
	
	/**
	 * @return list of descendants of this reading tree node
	 */
	/*
	 * test:
	 *   no sons
	 *   two sons, one with one or more sons
	 */
	public List getDescendants() {
		List d = new ArrayList();
		Iterator sit = this.getSons().iterator();
		while (sit.hasNext()) {
			ReadingTreeNode s = (ReadingTreeNode) sit.next();
			d.add(s);
			d.addAll(s.getDescendants());
		}
		return d;
	}
	

	/**
	 * @param l = info string of a possible ancestor of this node
	 * @return true iff l is the info string of an ancestor of this reading tree node
	 */
	/*
	 * test:
	 *   this node has no father
	 *   this node has father l
	 *   this node has grandfather l
	 *   this node has father and grandfather, but not l
	 */
	public boolean hasAncestor(String l) {
		ReadingTreeNode f = this.getFather();
		if (f == null) return false;
		return (f.getInfo().equals(l) || f.hasAncestor(l));
	}
	
	
	/**
	 * @return a string representation of this reading tree node
	 */
	public String toString() {
		String res = "";
		res += "("+info;
		for(int i=0; i<sons.size(); i++) {
			ReadingTreeNode son = (ReadingTreeNode)sons.get(i);
			res += son.toString();
		}
		res += ")";
		return res;
	}
	
	/**
	 * 
	 * @param l highest level of reading tree to be returned
	 * @return a string representation of the levels 1..l of this reading tree
	 */
	public String toString(int l) {
		String res = "";
		for(int i=0; i<l; i++) {
			res += indent;
		}	
		res += info+"\n";
		for(int i=0; i<sons.size(); i++) {
			ReadingTreeNode son = (ReadingTreeNode)sons.get(i);
			res += son.toString(l+1);
		}
		return res;
	}
	
	// operations
	// ----------

	/**
	 * Post: this node has been given the father node f
	 * @param f father node of this reading tree node
	 */
	public void setFather(ReadingTreeNode f) {
		this.father = f;
	}
	
	/**
	 * Post: this node has been given the son s.
	 *       S has been given this node as father
	 * @param s node of the reading tree
	 */
	public void addSon(ReadingTreeNode s) {
		this.sons.add(s);
		s.setFather(this);
	}
	
	//
	// Local properties
	//
	protected String info = "";
	protected List sons = new ArrayList();
	protected ReadingTreeNode father = null;
	protected String indent = " ";
	
	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	
    	{
    		hd("Test of attributes");
    		ReadingTreeNode fred = new ReadingTreeNode("fred");
    		pr("fred="+fred);
    		pr("fred.getInfo()="+fred.getInfo());
    		pr("fred.getFather()="+fred.getFather());
    		pr("fred.getSons()="+fred.getSons());
    		pg();
    		
    		
    		ReadingTreeNode jim = new ReadingTreeNode("jim");
    		fred.addSon(jim);
    		pr("fred="+fred);
    		pr("fred.getInfo()="+fred.getInfo());
    		pr("fred.getFather()="+fred.getFather());
    		pr("fred.getSons()="+fred.getSons());
    		pr("jim="+jim);
    		pr("jim.getInfo()="+jim.getInfo());
    		pr("jim.getFather()="+jim.getFather());
    		pr("jim.getSons()="+jim.getSons());
    		pg();
    	}
    	
    	
    	{
    		hd("Tests of getDescendants");
    		ReadingTreeNode fred = new ReadingTreeNode("fred");
    		pr("fred="+fred);
    		pr("fred.getDescendants()="+fred.getDescendants());
    		pg();
    		
    		ReadingTreeNode jim = new ReadingTreeNode("jim");
    		fred.addSon(jim);
    		
    		ReadingTreeNode joe = new ReadingTreeNode("joe");
    		fred.addSon(joe);
    		
    		ReadingTreeNode jack = new ReadingTreeNode("jack");
    		jim.addSon(jack);
    		pr("fred="+fred);
    		pr("jim="+jim);
    		pr("joe="+joe);
    		pr("jack="+jack);
    		pr("fred.getDescendants()="+fred.getDescendants());
       		pg();
    	}
    	
    	{
    		hd("Tests of hasAncestor");
    		ReadingTreeNode fred = new ReadingTreeNode("fred");
    		pr("fred="+fred);
    		pr("fred.getInfo()="+fred.getInfo());
    		pr("fred.getFather()="+fred.getFather());
    		pr("fred.getSons()="+fred.getSons());
    		pg();
    		
    		pr("fred.hasAncestor(\"jim\")="+fred.hasAncestor("jim"));
    		pg();
    		
    		ReadingTreeNode jim = new ReadingTreeNode("jim");
    		ReadingTreeNode joe = new ReadingTreeNode("joe");
    		fred.addSon(jim);
    		jim.addSon(joe);
    		pr("fred="+fred);
    		pr("fred.getInfo()="+fred.getInfo());
    		pr("fred.getFather()="+fred.getFather());
    		pr("fred.getSons()="+fred.getSons());
    		pg();

    		pr("joe.hasAncestor(\"jim\")="+joe.hasAncestor("jim"));
    		pg();
    		
    		pr("joe.hasAncestor(\"fred\")="+joe.hasAncestor("fred"));
    		pg();
    		
    		pr("joe.hasAncestor(\"bert\")="+joe.hasAncestor("bert"));
    		pg();
    		
    	}
    }
    */
}
