package rwd.ocl.constraintdiagram;

/**
 * A concrete string representation of a reading tree.
 * The Plane Tiling Condition node, which is the root, is not represented.
 * Other nodes are represented by their label names (IDs).
 * A tree is represented by an expression in parentheses.
 * The expression consists of the root node ID and a forest representing the 
 * children of that node.
 * Syntax:
 * <PRE>
     readingTreeString ->  forest 
     forest -> tree*
     tree -> '(' ID forest ')' 
   </PRE>
 * @author Richard Bosworth
 * 
 * @version 1.8.05  10:18
 */
public class ReadingTreeString {

	String myString;  // string of this reading tree
	Token t;          // current token of string
	
	
	// constructor
	// -----------
	
	/**
	 * post: a ReadingTreeString object has been constructed for s.
	 * @param s string representing a valid reading tree
	 */
	public ReadingTreeString(String s) {
		myString = s;
	}
	

	// attributes
	// ----------
	
	/**
	 * @return string representing label of last (rightmost) node in the 
	 *           reading tree string, or the empty string if none
	 */
	/*
	 * test:
	 *   reading tree with no nodes
	 *   reading tree with two nodes
	 */
	public String lastNode(){
		String result = "";
		Lexer l = new StringLexer(myString);
		while (true) {
			t = l.nextToken();
			if (t.code == Token.END) break;
			if (t.code == Token.ID)
				result = t.info;
		}
		return result;
	}
	
	
	/**
	 * @return a string representation of this ReadingTreeString.
	 */
	/*
	 * test: once thru
	 */
	public String toString() {
		return myString;
	}
	
	// operations
	// ----------
	
	/**
	 * post: y has been attached to the first occurrence of x in the string
	 * @param x label of node to which new node is to be attached as child. 
	 * @param y label of new node to be attached.
	 */
	/*
	 * test: 
	 *   x is ""
	 *     null string
	 *     string with forest
	 *     string with forest and ")"
	 *   x is non-null
	 *     no x in string
	 *     x in string
	 *       x has no children
	 *       x has two or more children
	 */
	public void addNode(String x, String y) {
		
		// build string with y attached to x as a child
		String result = "";
		Lexer l = new StringLexer(myString);
		if (x.equals("")) {
			
			// skip any initial forest 
			result = skipForest(l,result);
			
			// add y as child of PTC
			result += "(" + y + ")";
			
			// add current token, if any
			if (t.code != Token.END) 
				result += t.info;

			myString = result;
			return;
		}
		while (true) {
			t = l.nextToken();
			if (t.code == Token.END) break;
			if (t.code == Token.ID && t.info.equals(x)) {
				result += t.info;
				
				// skip the forest after x
				result = skipForest(l,result);
				
				// add y as child
				result += "(" + y + ")";
				
				// add current token, if any
				if (t.code != Token.END) 
					result += t.info;
			}
			else 
				result += t.info;
		}
		
		// replace myString with result
		myString = result;
	}
	
	
	/**
	 * post: the first occurrence of x in the string has been replaced by y
	 * @param x old label of node to be modified
	 * @param y new label of node to be modified
	 */
	/*
	 * test:
	 *   x not present
	 *   x present
	 *     string with other tokens than x
	 */
	public void modifyNode(String x, String y) {
		// build string with x replaced by y
		String result = "";
		Lexer l = new StringLexer(myString);
		while (true) {
			t = l.nextToken();
			if (t.code == Token.END) break;
			if (t.code == Token.ID && t.info.equals(x))
				result += y;
			else 
				result += t.info;
		}
		
		// replace myString with result
		myString = result;
	}
	
	
	/**
	 * Post: the first occurrence of the node has been deleted,
	 *         and its children have been assigned to its father
	 * @param x id of node to be deleted
	 */
	/*
	 * test:
	 *   string without x
	 *   string with x having children and siblings of x
	 */
	public void deleteNode(String x) {
		String result = "";
		Lexer l = new StringLexer(myString);
		while (true) {
			t = l.nextToken();
			if (t.code == Token.END) break;
			if (t.code == Token.LP) {
				t = l.nextToken();
				if (t.code == Token.ID && t.info.equals(x)) {
					// ignore this node and its preceding left parenthesis
					result = skipForest(l,result);
					// current token is matching right parenthesis, which is ignored!
				}
				else {
					result += "(" + t.info;
				}
			}
			else
				result += t.info;
		}
		
		// replace myString with result
		myString = result;
	}
	
	/**
	 * Post: 
	 * <UL>
	 *   <LI>if oldlab is empty and newlab is not, adds new node with label 
	 *         newlab as child of last node in tree
	 *   <LI>if newlab is empty and oldlab is not, attempts to delete node 
	 *         with label oldlab
	 *   <LI>if newlab and oldlab are non-empty, attempts to modify oldlab's
	 *         label to newlab
	 *   <LI>if both oldlab and newlab are empty, does nothing
	 * </UL>
	 *       
	 * @param oldlab old value of node label (may be empty)
	 * @param newlab new value of node label (may be empty)
	 */
	/*
	 * test:
	 *   oldlab empty
	 *     newlab empty
	 *     newlab non-empty
	 *   oldlab non-empty
	 *     newlab empty
	 *     newlab non-empty
	 */
	public void update(String oldlab, String newlab) {
		if (oldlab.equals("")) {
			if (newlab.equals("")) 
				return;
			else
				addNode(lastNode(),newlab);
		} else /* oldlab not empty */ {
			if (newlab.equals(""))
				deleteNode(oldlab);
			else
				modifyNode(oldlab,newlab);
		}
	}
	
	
	// private methods
	// ---------------
	
	/*
	 * pre:  l = current lexer
	 *       r = result string for scan
	 * post: if l was headed by a forest,
	 *         returns r with the forest appended
	 *       else returns r
	 *       t = token after the forest
	 */
	/*
	 * test: 
	 *   null lexer
	 *   lexer with "fred"
	 *   lexer with ")"
	 *   lexer with "(fred) (jim (bert)) lionel"
	 */
	private String skipForest(Lexer l, String r){
		String result = r;
		int nest = 0;
		while (true) {
			t = l.nextToken();
			if (t.code == Token.END) break;
			if (t.code == Token.LP) {
				nest++;
				result += t.info;
			} else if (t.code == Token.RP) {
				if (nest == 0) break;
				nest--;
				result += t.info;
				
			} else {
				if (nest == 0) break;
				result += t.info;
			}
		}
		return result;
		
	}

	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	hd("Test of constructor");
    	ReadingTreeString rts = new ReadingTreeString("(fred)");
    	pr("rts="+rts);
    	pg();
    	/*
    	{
	    	hd("test of modify");
	    	pr("rts.modifyNode(\"jim\",\"bert\")");
	    	    rts.modifyNode("jim","bert");
	    	pr("rts="+rts);
	    	pg();
	    	
	    	pr("rts.modifyNode(\"fred\",\"fred\")");
	    	    rts.modifyNode("fred","fred");
	    	pr("rts="+rts);
	    	pg();
	    	pr("rts.modifyNode(\"fred\",\"jim\")");
	    		rts.modifyNode("fred","jim");
	    	pr("rts="+rts);
	    	pg();
    	}
    	
    	{
    		hd("Tests of skipForest");
    		String s = "";
    		Lexer l = new StringLexer(s);
    		String r = "a";
    		pr("s="+s);
    		pr("r="+r);
    		pr("rts.skipForest(l,r)="+rts.skipForest(l,r));
    		pg();

    		s = "fred";
    		l = new StringLexer(s);
    		r = "b";
    		pr("s="+s);
    		pr("r="+r);
    		pr("rts.skipForest(l,r)="+rts.skipForest(l,r));
    		pg();

    		s = ")";
    		l = new StringLexer(s);
    		r = "c";
    		pr("s="+s);
    		pr("r="+r);
    		pr("rts.skipForest(l,r)="+rts.skipForest(l,r));
    		pg();

    		s = "(fred) (jim (bert)) lionel";
    		l = new StringLexer(s);
    		r = "d";
    		pr("s="+s);
    		pr("r="+r);
    		pr("rts.skipForest(l,r)="+rts.skipForest(l,r));
    		pg();

    	}
    	
    	{
    		hd("Tests of addNode");
    		rts = new ReadingTreeString("");
    		pr("rts="+rts);
    		pr("rts.addNode(\"\",\"jim\")");
 		    rts.addNode("","jim");
    		pr("rts="+rts);
    		pg();
    		
    		rts = new ReadingTreeString("(fred) (boris)");
    		pr("rts="+rts);
    		pr("rts.addNode(\"\",\"jim\")");
 		    rts.addNode("","jim");
    		pr("rts="+rts);
    		pg();
    		
    		rts = new ReadingTreeString("(fred) (boris) )");
    		pr("rts="+rts);
    		pr("rts.addNode(\"\",\"jim\")");
 		    rts.addNode("","jim");
    		pr("rts="+rts);
    		pg();
    		
    		rts = new ReadingTreeString("(fred)");
    		pr("rts="+rts);
     		pr("rts.addNode(\"freda\",\"bert\")");
 		    rts.addNode("freda","bert");
    		pr("rts="+rts);
    		pg();
    		
    		pr("rts.addNode(\"fred\",\"bert\")");
 		    rts.addNode("fred","bert");
    		pr("rts="+rts);
    		pg();
    		
    		pr("rts.addNode(\"fred\",\"lionel\")");
 		    rts.addNode("fred","lionel");
    		pr("rts="+rts);
    		pg();
    		
    		pr("rts.addNode(\"fred\",\"bob\")");
 		    rts.addNode("fred","bob");
    		pr("rts="+rts);
    		pg();
    	}
    	
    	{
    		hd("Tests of deleteNode");
    		rts = new ReadingTreeString(" ( millie ) ( fred ( bob ) ( nellie ) ) ( jim ) ( bert ) * (");
    		pr("rts="+rts);
    		pr("rts.deleteNode(\"lionel\")");
    		rts.deleteNode("lionel");
    		pr("rts="+rts);
    		pg();
    		
    		pr("rts.deleteNode(\"fred\")");
    		rts.deleteNode("fred");
    		pr("rts="+rts);
    		pg();
    	}
    	
    	
    	{
    		hd("Tests of lastNode");
    		rts = new ReadingTreeString("");
    		pr("rts="+rts);
    		pr("rts.lastNode()="+rts.lastNode());
    		pg();
    		
    		rts = new ReadingTreeString("(esmeralda(felicity))");
    		pr("rts="+rts);
    		pr("rts.lastNode()="+rts.lastNode());
    		pg();
    		
    		
    		
    	}
    	
    	{
    		hd("Test of update");
    		pr("rts="+rts);
    		pr("rts.update(\"\",\"\")");
    		rts.update("","");
    		pr("rts="+rts);
    		pg();
    		
    		pr("rts.update(\"\",\"bert\")");
    		rts.update("","bert");
    		pr("rts="+rts);
    		pg();
    		
    		pr("rts.update(\"fred\",\"jim\")");
    		rts.update("fred","jim");
    		pr("rts="+rts);
    		pg();
    		
    		pr("rts.update(\"bert\",\"\")");
    		rts.update("bert","");
    		pr("rts="+rts);
    		pg();
    		
    		
    	}
    }
    */
}
