package rwd.ocl.eulerdiagram;

import java.util.Vector;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;

import uk.ac.kent.cs.kmf.common.Repository;
import uk.ac.kent.cs.kmf.util.XMIToUMLLoader;
import uk.ac.kent.cs.kmf.util.ILog;
import uk.ac.kent.cs.kmf.util.Pair;
import uk.ac.ukc.cs.kmf.kmfstudio.uml.Foundation.Core.Class_;
import uk.ac.ukc.cs.kmf.kmfstudio.uml.Foundation.Core.Classifier;
import uk.ac.ukc.cs.kmf.kmfstudio.uml.Foundation.Core.Generalization;
import uk.ac.ukc.cs.kmf.kmfstudio.uml.Model_Management.Model;
import uk.ac.ukc.cs.kmf.kmfstudio.uml.repository.UmlRepository;
import uk.ac.kent.cs.kmf.util.Naming;

/**
 * @author Octavian Patrascoiu
 *
 */
public class Uml2EulerDiagram {
	/** Constructor */
	public Uml2EulerDiagram(ILog log) {
		this.log = log;
	}
	
	/** Generate the spider diagram */
	public UnitaryDiagram makeEulerDigram(String xmiFileName) {
		// Load UML model
		this.xmiFileName = xmiFileName;
		loadModel(xmiFileName, log);
		// Compute all the UML Classes
		List classes = getUMLClasses();
			
		// Compute contours and zones
		Vector contours = makeContours(classes);
		Vector zones = makeZones(classes, contours);
		
		// Create the unitary diagram
		UnitaryDiagram ud = new UnitaryDiagram(new Vector(zones), "ud");
		
		// Print result
		log.reportMessage(ud.toDebugString());

		return ud;			
	}
	
	/** Compute all the contours: classes which does not inherit from nother */
	protected Vector makeContours(List classes) {
		// Create a contour for each class which does not inherit from another
		Vector contours = new Vector();
		for(int i=0; i<classes.size(); i++) {
			Class_ cls = (Class_)classes.get(i);
			if (!isOuterZone(cls)) {
				// Compute direct super classes
				Set supers = cls.getGeneralization();
				// There is no super class 
				if (supers.size() == 0) {
					// Make a contour
					Contour cont = Contour.getContour(Naming.getFullClassifierName(cls));
					// Add contour
					if (!contours.contains(cont)) contours.add(cont);
				// There is only one super class 
				} else if (supers.size() == 1) {
					// Make a contour
					Contour cont = Contour.getContour(Naming.getFullClassifierName(cls));
					// Add contour
					if (!contours.contains(cont)) contours.add(cont);
				// Else
				} else {
				}
			}
		}
		return contours;
	}

	/** Compute all the contours: classes which does not inherit from nother */
	protected Vector makeZones(List classes, Vector contours) {
		// Create a contour for each class which does not inherit from another
		Vector zones = new Vector();
		for(int i=0; i<classes.size(); i++) {
			Class_ self = (Class_)classes.get(i);
			String selfName = Naming.getFullClassifierName(self);
			if (!isOuterZone(self)) {
				// Compute all super classes
				Iterator it = self.getGeneralization().iterator();
				List supers = new Vector();
				while (it.hasNext()) {
					Generalization gen = (Generalization)it.next();
					Object parent = gen.getParent();
					if (parent != null) {
						supers.add(parent);
					}
				}
				// There is no super class 
				if (supers.size() == 0) {
					// Make a zone
					Vector in = new Vector();
					in.add(Contour.getContour(selfName));
					Vector out = new Vector((Vector)contours.clone());
					out.removeAll(in);
					Zone zone = Zone.getZone(in, out);
					// Add zone
					if (!zones.contains(zone)) zones.add(zone);
				// There is only one super class
				} else if (supers.size() == 1) {			
					// Make a zone
					Vector in = new Vector();
					in.add(Contour.getContour(selfName));
					Class_ sup = (Class_)supers.get(0);
					String supName = Naming.getFullClassifierName(sup);
					in.add(Contour.getContour(supName));
					Vector out = new Vector((Vector)contours.clone());
					out.removeAll(in);
					Zone zone = Zone.getZone(in, out);
					// Add zone
					if (!zones.contains(zone)) zones.add(zone);			
				// There are at least 2 super classes
				} else {
					// Make a zone
					Vector in = new Vector();
					Iterator j=supers.iterator(); 
					while (j.hasNext()) {
						Class_ sup = (Class_)j.next();
						in.add(Contour.getContour(Naming.getFullClassifierName(sup)));
					}
					Vector out = new Vector((Vector)contours.clone());
					out.removeAll(in);
					Zone zone = Zone.getZone(in, out);
					// Add zone
					if (!zones.contains(zone)) zones.add(zone);
				}
			}
		}

		// Add outer zone
		Zone outer = Zone.getZone(new Vector(), new Vector(contours));
		zones.add(outer);
		
		return zones;
	}
	
	/** Check if is the outer zone */
	protected boolean isOuterZone(Class_ cls) {
		return cls.getName().getBody_().equals("Out"); 
	}
		
	/** Input XMI file name */
	private String xmiFileName;
	protected String getXmiFileName() {
		return xmiFileName;
	}
	protected void setXmiFileName(String string) {
		xmiFileName = string;
	}

	/** UML Model */
	private Model model;
	protected Model getModel() {
		return model;
	}
	protected void setModel(Model model) {
		this.model = model;
	}

	/** UML repository */
	private Repository rep;
	protected Repository getRep() {
		return rep;
	}
	protected void setRep(UmlRepository repository) {
		rep = repository;
	}
		
	/** Init model and UML repository */
	protected void loadModel(String xmiFileName, ILog log) {
		Pair pair = null;
		try {
			//--- Load the model ---
			pair = (new XMIToUMLLoader()).loadModel(xmiFileName, log); 
			model = (Model)pair.getFirst();
			rep = (Repository)pair.getSecond();
		} catch (Exception e) {
			log.reportError("XML error "+e.toString());
		}
	}

	/** Compute all the UML classes */
	protected List getUMLClasses() {
		if (rep == null) loadModel(xmiFileName, log);
		return rep.getElements("uml.Foundation.Core.Class_");
	}

	/** Look for a UML class with a given name */
	protected Class_ getClass(List classes, String clsName) {
		clsName = clsName.replaceAll("::", ".");
		for(int i=0; i<classes.size(); i++) {
			Class_ cls = (Class_)classes.get(i);
			String name = Naming.getFullClassifierName(cls);
			if (name.equals(clsName)) {
				 return cls;
			}
		}
		return null;
	}

	/** Remove super classes which are not used */
	protected List getSuperClasses(Class_ cls, List usedClasses) {
		List usedSuperCls = new Vector();
		List superCls = Naming.allSuperClasses(cls, false);
		superCls.remove(cls);
		for(int i=0; i<superCls.size(); i++) {
			Classifier sup = (Classifier)superCls.get(i);
			String superName = Naming.getFullClassifierName(sup);
			if (usedClasses.contains(superName)) {
				usedSuperCls.add(superName);	
			}
		}
		return usedSuperCls; 
	}

	/** Log */
	private ILog log;
	public ILog getLog() {
		return log;
	}
	public void setLog(ILog log) {
		this.log = log;
	}
}
