package rwd.ocl.spiderdiagram;

import java.util.*;

import rwd.abstractsyntax.eulerdiagram.*;
import rwd.abstractsyntax.spiderdiagram.*;

import rwd.concretesyntax.*;
import rwd.concretesyntax.eulerdiagram.*;

import rwd.ocl.eulerdiagram.EulerDiagram2Uml;
import rwd.tableaux.diagram.*;
import rwd.tableaux.semantic.OptionsImpl;
import rwd.tableaux.syntax.ast.*;
import rwd.tableaux.syntax.parser.*;

import uk.ac.kent.cs.kmf.util.ILog;

/**
 * @author Octavian Patrascoiu
 *
 */
public class SpiderDiagram2Ocl {
	/** Constructor */
	public SpiderDiagram2Ocl(List diagrams, ILog log) {
		// Initialize input members
		this.cds = new LinkedHashSet(diagrams);
		this.log = log;
		// Compute cuds, cd, and ud
		this.cuds = new LinkedHashSet();
		Iterator i = cds.iterator();
		while (i.hasNext()) {
			ConcreteDiagram cd = (ConcreteDiagram)i.next();
			Set cuds = cd.getConcreteUnitaryDiagrams();
			this.cuds.addAll(cuds);
		}
	}

	/** Debug/release flag */
	public boolean isDebug() {
		return debug;
	}
	public void setDebug(boolean b) {
		debug = b;
	}

	/** Generate the equivalent OCL */
	public String makeOCL(String stringExp) {
		LogicalExp exp = parser.parse(stringExp, log);
		String oclStr = getOCL(exp);
		if (oclStr == null || oclStr.length() == 0) {
			oclStr = "true";
		} 
		return "context OclVoid inv: "+oclStr;
	}
	/** Compute the OCL string for a logical expression */
	protected String getOCL(LogicalExp exp) {
		if (exp instanceof AtomicExp) {
			String name = ((AtomicExp)exp).getName();
			// Simple cases
			if (name.equals("true")) return "true";
			else if (name.equals("false")) return "false";
			else return getOCL4UnitaryDiagram(name);
		} else if (exp instanceof NotExp) {
			return " not "+getOCL(exp.getLeft());
		} else {
			// Generate for left and right Add left and right expressions
			LogicalExp leftExp = exp.getLeft();
			LogicalExp rightExp = exp.getRight();
			String leftStr = getOCL(leftExp);
			String rightStr = getOCL(rightExp);
			// P and Q goes to P, Q
			if (exp instanceof AndExp) 
				return leftStr + " and " + rightStr;
			else if (exp instanceof OrExp) 
				return leftStr + " or " + rightStr;
			else if (exp instanceof ImpliesExp) 
				return leftStr + ".not() or " + rightStr;
			else if (exp instanceof EquivalentExp) 
				return "("+leftStr + ".not() or " + rightStr + ") and ("+leftStr + " or " + rightStr + ".not())";
			return null;
		}
	}
	/** Compute the OCL string for a unitary diagram with a given name */
	protected String getOCL4UnitaryDiagram(String name) {
		// Compute cd, cud, and all the concreteZones
		ConcreteDiagram cd = null;
		ConcreteUnitaryDiagram cud = null;
		// Compute cuds, cd, and ud
		Iterator i = cds.iterator();
		while (i.hasNext()) {
			ConcreteDiagram cdTemp = (ConcreteDiagram)i.next();
			Set cuds = cdTemp.getConcreteUnitaryDiagrams();
			this.cuds.addAll(cuds);
			Iterator j = cuds.iterator();
			ConcreteUnitaryDiagram cudTemp = null;
			while (j.hasNext()) {
				cudTemp = (ConcreteUnitaryDiagram)j.next();
				if (cudTemp.getName().equals(name)) {
					cd = cdTemp;
					cud = cudTemp;
					break;
				}
			}
		}
		// Check if the diagram exists
		if (cd == null || cud == null) {
			log.reportError("Cannot find diagram '"+name+"'");
			return null;
		}
		
		// Compute ocl
		List vars = new Vector();
		List exps = new Vector();
		// Compute all concrete zones
		Set allConcreteZones = cud.makeAllConcreteZones();
		// Compute variables
		vars = getFullVars(allConcreteZones);
		// Compute body
		exps = getFullBodyExp(cud);
		
		// Compute invariant
		String ocl = "let \n";
		ocl += "\t"+makeString(vars, ",\n\t")+"\n";
		ocl += "in\n";
		String body = makeString(exps, " or \n\t");
		if (body == null || body.length() == 0) body = "true"; 
		ocl += "\t"+body+"\n";
		return ocl;
	}

	/** Compute all the variables */
	protected List getFullVars(Set allConcreteZones) {
		Set vars = new LinkedHashSet();
		// Generate variable for each zone
		Collection zones = allConcreteZones;
		Set contours = new TreeSet();
		Iterator i = zones.iterator();
		while (i.hasNext()) {
			ConcreteZone zone = (ConcreteZone)i.next();
			contours.addAll(zone.getAbstractZone().getAllContours());
			// Compute the corresponding set of elements
			String varName = EulerDiagram2Uml.getVarName(zone);
			String init = EulerDiagram2Uml.getInitValue(zone);
			vars.add(varName+": Set(OclAny) = "+init);
		}
		// Add out variable
		String initOut = EulerDiagram2Uml.getInitValue4Out(new Vector(contours));
		vars.add("out :Set(OclAny) = "+initOut);
		return new Vector(vars);
	}


	/** Compute the body of the invariant for a concrete unitary diagram */
	protected List getFullBodyExp(ConcreteUnitaryDiagram cud) {
		// Split spiders
		SpiderSplitter splitter = new SpiderSplitterImpl(new OptionsImpl(), log);
		UnitaryDiagram aud = ConcreteUnitaryConverter.makeAUD(cud);
		Diagram d = splitter.splitSpiders(aud);
		// Get list of unitary diagrams
		List auds = d.getUnitaryParts();
		List exps = new Vector();
		// Make an or of ands
		for(int i=0; i<auds.size(); i++) {
			String andExp = getBodyExp((UnitaryDiagram)auds.get(i));
			if (andExp.length() != 0) 
				exps.add(andExp);
		}
		return exps;
	}
	protected String getBodyExp(UnitaryDiagram d) {
		List zones = d.getCloneOfZoneSet();
		List sZones = d.getCloneOfShadedZones();
		List spiders = d.getSpiders();
		List result = new Vector();
		for(int i=0; i<zones.size(); i++) {
			Zone az = (Zone)zones.get(i);
			// Compute number of feet and isShaded
			int feetNo = 0;
			for(int j=0; j<spiders.size(); j++) {
				Spider spider = (Spider)spiders.get(j);
				Collection habitat = spider.getHabitat();
				if (habitat.contains(az))
					feetNo++;
			}
			boolean isShaded = sZones.contains(az);
			// Compute expression
			String varName = EulerDiagram2Uml.getVarName(az);
			String exp = "(";
			if (isShaded) {
				if (feetNo == 0) exp += varName+"->size() = 0";
				else exp += varName+"->size() = "+feetNo;
			} else {
				if (feetNo == 0);
				else exp += varName+"->size() >= "+feetNo;
			}
			exp += ")";
			if (!exp.equals("()")) {
				result.add(exp);
			}
		}
		String str = makeString(result, " and ");
		return str;	
	}

	/** Make a string from a collection using the delimiter */
	protected String makeString(Collection objs, String del) {
		String res = "";
		Iterator i = objs.iterator();
		while (i.hasNext()) {
			Object obj = i.next();
			res += obj;
			if (i.hasNext()) res += del;
		}
		return res;		
	}

	// Members
	/** Debug/release flag */
	protected boolean debug = false;
	/** Logical expression parser */
	protected Parser parser = new ParserImpl();
	/** Concrete diagrams */
	protected Set cds;
	/** Concrete unitary diagrams */
	protected Set cuds;
	/** Log */
	protected ILog log;
	/** Display meaning for missing zones */
	protected static boolean MEANING_FOR_MISSING_ZONES = true;
}
