package rwd.rules.constraintdiagrams;

import java.awt.Color;
//import java.awt.Polygon;
//import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
//import java.util.List;
import java.util.Set;
//import java.util.Vector;

//import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
//import rwd.abstractsyntax.spiderdiagram.Diagram;
import rwd.abstractsyntax.constraintdiagram.Arrow;
//import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
//import rwd.concretesyntax.ConcreteDiagram;
import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.concretesyntax.DiagramItem;
//import rwd.concretesyntax.DiagramItem;
import rwd.concretesyntax.constraintdiagram.ConcreteArc;
import rwd.concretesyntax.constraintdiagram.ConcreteArcEnd;
import rwd.concretesyntax.eulerdiagram.ConcreteContour;
import rwd.concretesyntax.eulerdiagram.ConcreteZone;
import rwd.concretesyntax.eulerdiagram.DecorationBox;
//import rwd.concretesyntax.eulerdiagram.DecorationBox;
import rwd.concretesyntax.spiderdiagram.ConcreteSpider;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderFoot;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderLeg;
//import rwd.concretesyntax.spiderdiagram.ConcreteSpiderFoot;
//import rwd.concretesyntax.spiderdiagram.ConcreteSpiderLeg;
//import rwd.concretesyntax.eulerdiagram.ConcreteZoneSet;

/**
 * Reconciles a concrete diagram with its abstract equivalent.
 * A Reconciler is used after a rule has been applied at the abstract level, 
 * to create a concrete diagram which mirrors the resultant abstract diagram.
 *  
 * @author Richard Bosworth
 * 
 * @version 5.9.05   12:19
 */

public class Reconciler {

	// commands
	// --------
	
	/**
	 * Reconciles a ConcreteUnitaryDiagram with an abstract UnitaryDiagram 
	 * to produce a new ConcreteUnitaryDiagram.
	 * Some items which are in the ConcreteUnitaryDiagram and not in the 
	 * abstract UnitaryDiagram may be deleted from the ConcreteUnitaryDiagram.
	 * Items which are in the abstract UnitaryDiagram and not in the 
	 * ConcreteUnitaryDiagram may have a concrete representation made, 
	 * which is then added to the ConcreteUnitaryDiagram.
	 * Items which are common to both diagrams may be modified in the 
	 * ConcreteUnitaryDiagram.
	 * 
	 * @param cud ConcreteUnitaryDiagram to be reconciled
	 * @param ud abstract UnitaryDiagram to be reconciled
	 * @return ConcreteUnitaryDiagram which is the reconciliation of cud and ud
	 */
	/*
	 * tests:
	 *   concrete diagram with decoration box d and contour a
	 *     and zones ((ad,),shaded)  ((d,a),unshaded)
	 *     abstract diagram with zones ((a,):unshaded) ((,a):shaded)
	 *       no spiders in the concrete diagram
	 *       at least two spiders in the concrete diagram
	 *       no arcs in the concrete diagram
	 *       at least two arcs in the concrete diagram
	 *         arc from nowhere to nowhere
	 *         arc from contour to contour
	 * 		   arc from spider foot to spider foot
	 * 
	 */
	public ConcreteUnitaryDiagram reconcile(ConcreteUnitaryDiagram cud, 
			                                UnitaryDiagram ud) {

		//Set up map from original concrete diagram to new concrete diagram
		HashMap origToNew = new HashMap();
		
		// create diagram newcud with decoration box of cud
		DecorationBox db = cud.getDecorationBox();
		DecorationBox dbc = (DecorationBox) db.getClone();
		ConcreteUnitaryDiagram newcud = new ConcreteUnitaryDiagram(dbc);
		origToNew.put(db,dbc);
		ResultSets rs; 

		// contours
		// --------
		
		// intersect cud contours with ud contours		
		rs = intersectContours(cud.getConcreteContours(),
				               ud.getCloneOfContourSet());
		// add common contour clones to newcud
		for (Iterator cit = rs.common.iterator(); cit.hasNext();) {
			ConcreteContour cc = (ConcreteContour) cit.next();
			ConcreteContour ccc = (ConcreteContour) cc.getClone();
			newcud.addConcreteContour(ccc);
			origToNew.put(cc,ccc);
		}
		
		// zones
		// -----
   		// shade zones as necessary from ud
   		Collection czones = newcud.getConcreteZoneSet().getZones();
   		for (Iterator zit = czones.iterator(); zit.hasNext();) {
   			ConcreteZone cz = (ConcreteZone) zit.next();
   			ConcreteZone czc = (ConcreteZone) cz.getClone();
			if (cz.getAbstractZone().isShaded(ud))
				cz.setBackColor(Color.LIGHT_GRAY);
			else
				cz.setBackColor(Color.WHITE);
   		}
   		
		// spiders
		// -------
   		
		// intersect cud spiders with ud spiders
   		rs = intersectSpiders(cud.getConcreteSpiders(),
   				              ud.getSpiders(),
							  cud);
   		// add clones of feet and legs of common spiders to newcud
   		for (Iterator spit = rs.common.iterator(); spit.hasNext(); ) {
   			ConcreteSpider cs = (ConcreteSpider) spit.next();
  	   		
   	   		// add cloned feet of spider to newcud and to orig->new mapping
   	   		Iterator fit = cs.getFeet().iterator();
   	   		while (fit.hasNext()) {
   	   			ConcreteSpiderFoot csf = (ConcreteSpiderFoot) fit.next();
   	   			ConcreteSpiderFoot csfc = (ConcreteSpiderFoot) csf.getClone();
   	   			newcud.addConcreteFoot(csfc);
   	   			origToNew.put(csf,csfc);
   	   		}
   	   		
   	   		// add cloned legs of spider to newcud
   	   		Iterator lit = cs.getLegs().iterator();
   	   		while (lit.hasNext()) {
   	   			ConcreteSpiderLeg csl = (ConcreteSpiderLeg) lit.next();
   	   			// make clone of leg using already cloned feet
   	   			ConcreteSpiderFoot newSource = 
   	   				(ConcreteSpiderFoot) origToNew.get(csl.getSource());
   	   			ConcreteSpiderFoot newTarget = 
   	   				(ConcreteSpiderFoot) origToNew.get(csl.getTarget());
   	   			
   	   			ConcreteSpiderLeg cslc = 
   	   				new ConcreteSpiderLeg(csl.getLabel(),newSource,newTarget,csl.getType());
   	   			newcud.addConcreteLeg(cslc);
   	   		}
   		}
   	    		
		// arrows
		// ------
		// intersect cud arrows with ud arrows
   		rs = intersectArrows(cud.getConcreteArcs(),
		                     ud.getArrows(),
				             cud);
   		// add common arrows to newcud
   		for (Iterator arit = rs.common.iterator(); arit.hasNext(); ) {
   			ConcreteArc ca = (ConcreteArc) arit.next();
   			ConcreteArc cac = new ConcreteArc();
   			cac.setType(ca.getType());
   			
   			// find original attachments
   			ConcreteArcEnd start = ca.getStart();
   			ConcreteArcEnd middle = ca.getMiddle();
   			ConcreteArcEnd end = ca.getEnd();
   			DiagramItem source = ca.getSource();
   			DiagramItem target   = ca.getTarget();
   			
   			// set up arrow ends to their clones 
   			ConcreteArcEnd newStart;
   			ConcreteArcEnd newMiddle;
   			ConcreteArcEnd newEnd;
   
   			if (source == null) {
   				newStart = new ConcreteArcEnd(null);
   			}
   			else { 
   				DiagramItem newSource = (DiagramItem) origToNew.get(source);
   				newStart = new ConcreteArcEnd(newSource);
   				newStart.attachItem(newSource);
   			}
   			newStart.setLocation(start.getX(),start.getY());
   			
			newMiddle = new ConcreteArcEnd(null);
   			newMiddle.setLocation(middle.getX(),middle.getY());
   			newMiddle.setIsMiddle(true);
   			
   			if (target == null) {
   				newEnd = new ConcreteArcEnd(null);
   			}
   			else {
   				DiagramItem newTarget = (DiagramItem) origToNew.get(target);
   				newEnd = new ConcreteArcEnd(newTarget);
   				newEnd.attachItem(newTarget);
   			}
   			newEnd.setLocation(end.getX(),end.getY());

   			// attach cloned ends to the cloned arrow
   			cac.setStart(newStart);
   			cac.setMiddle(newMiddle);
   			cac.setEnd(newEnd);
   			
   			// set up label
   			cac.setLabel(ca.getLabel());
   			cac.setLabelPosition(ca.getLabelPosition()); 
   		   	
   			// add cloned arrow to new diagram
   	   		newcud.addConcreteArc(cac);
   		}
   		
   		// reading tree
   		// ------------
   		newcud.getDecorationBox().setReadingForest(ud.getReadingForest());
   		
   		return newcud;
	}
	
	
	/*
	 * Reconciles a ConcreteDiagram with an abstract Diagram to produce a 
	 *   new ConcreteDiagram.
	 * Some items which are in the ConcreteDiagram and not in the abstract 
	 * Diagram may be deleted from the ConcreteDiagram.
	 * Items which are in the abstract Diagram and not in the ConcreteDiagram 
	 * may have a concrete representation made, which is then added to the 
	 * ConcreteDiagram.
	 * Items which are common to both diagrams may be modified in the 
	 * ConcreteDiagram.
	 * 
	 * @param cd ConcreteDiagram to be reconciled
	 * @param ad abstract Diagram to be reconciled
	 * @return ConcreteDiagram which is the reconciliation of cd and ad
	 */
	//public ConcreteDiagram reconcile(ConcreteDiagram cd, Diagram ad) {
	//	return cd; // dummy "do nothing" for now
	//}
	
	
	// private methods
	// ---------------
	
	/*
	 * inner class used to pass out the result of intersecting two sets
	 */
	private class ResultSets {
        Set cdonly;  // items that are only in the concrete diagram
        Set common;  // items that are in both concrete diagram and abstract diagram
        Set adonly;  // items that are only in the abstract diagram
        
    	// queries
    	// -------
    	
		/*
		post: returns a string representation of the class
		*/
		/*
		test: once thru
		*/
		public String toString() {
		    return super.toString() + "[" + toStringContents() + "]";
		}
		  
		/*
		post: returns a string representation of the class contents
		*/
		/*
		test: once thru
		*/
		public String toStringContents() {
		    return "cdonly=" + cdonly +
		          ",common=" + common +
		          ",adonly=" + adonly;
		}
	}

	/*
	 * pre:  cditems = contours of a concrete diagram
	 *       aditems = contours of an abstract diagram
	 * post: returns resultSets giving intersection and symmetric difference
	 *       (note: intersection has concrete contours)
	 */
	/*
	 * test:
	 *   null cditems
	 *     null aditems
	 *     aditems has one contour
	 *   cditems has contours a,b,c 
	 *     aditems has contours b,c,d
	 */
	private ResultSets intersectContours(Set cditems, Collection aditems) {
		ResultSets rs = new ResultSets();
		rs.cdonly = cditems;
		rs.adonly = new HashSet(aditems);
		rs.common = new HashSet();
		
		Iterator it = rs.cdonly.iterator();
		while (it.hasNext()) {
			ConcreteContour cc = (ConcreteContour) it.next();
			if (rs.adonly.contains(cc.getContour())){
				rs.common.add(cc);
				it.remove();
				rs.adonly.remove(cc.getContour());
			}
		}
		
		return rs;
	}
	
	/*
	 * pre:  cditems = spiders of a concrete diagram
	 *       aditems = spiders of an abstract diagram
	 *       cud = ConcreteUnitaryDiagram in which spiders appear
	 * post: returns resultSets giving intersection and symmetric difference
	 *       (note: intersection contains concrete spiders)
	 *       Any concrete spider in the intersection set whose feet are a 
	 *         subset of the feet of the matching abstract spider 
	 *         has the extra feet added
	 */
	/*
	 * test:
	 *   null cditems
	 *     null aditems
	 *     aditems has one spider
	 *   aditems has spiders with feet in zones: (x,y)(y,z) 
	 *     cditems has spiders with feet in zones: (y,z)(z,x) 
	 */
	private ResultSets intersectSpiders(Set cditems, 
			                            Collection aditems,
										ConcreteUnitaryDiagram cud) {
		ResultSets rs = new ResultSets();
		rs.cdonly = cditems;
		rs.adonly = new HashSet(aditems);
		rs.common = new HashSet();
		
		Iterator cit = rs.cdonly.iterator();
		while (cit.hasNext()) {
			ConcreteSpider csp = (ConcreteSpider) cit.next();
			Iterator ait = rs.adonly.iterator();
			while (ait.hasNext()){
				Spider adasp = (Spider)ait.next();
				if(match(adasp,csp,cud)){
					rs.common.add(csp);
					try{
						cit.remove();
					}
					catch(IllegalStateException e){}  // ignore attempts to delete spider twice
					ait.remove();
				}
			}
		}
		
	return rs;
	}
	
	
	/*
	 * pre:  cditems = arrows of a concrete diagram
	 *       aditems = arrows of an abstract diagram
	 *       cud = ConcreteUnitaryDiagram in which arrows appear
	 * post: returns resultSets giving intersection and symmetric difference
	 *       (note: intersection contains concrete arrows)
	 */
	/*
	 * test:
	 *   null cditems
	 *     null aditems
	 *     aditems has one arrow
	 *   aditems has arrows between (a,a) (a,b)
	 *     cditems has arrows between (a,b) (b,a) 
	 */
	private ResultSets intersectArrows(Set cditems, 
			                           Collection aditems,
									   ConcreteUnitaryDiagram cud) {
		ResultSets rs = new ResultSets();
		rs.cdonly = cditems;
		rs.adonly = new HashSet(aditems);
		rs.common = new HashSet();
		
		Iterator cit = rs.cdonly.iterator();
		while (cit.hasNext()) {
			ConcreteArc carrow = (ConcreteArc) cit.next();
			Arrow arrow = carrow.getAbstractArrow(cud);
			Iterator ait = rs.adonly.iterator();
			while (ait.hasNext()){
				Arrow adarrow = (Arrow)ait.next();
				if (arrow.equals(adarrow)){
					rs.common.add(carrow);
					cit.remove();
					ait.remove();
				}
			}
		}
		
	return rs;
	}
	
	/*
	 * pre:  s = abstract spider
	 *       cs = concrete spider
	 *       cud = concrete unitary diagram containing cs
	 * post: returns true iff the spiders match (the labels of the spiders are 
	 *         identical and the habitat of the concrete spider
	 *         is a subset of the habitat of the abstract spider).
	 *         Any extra legs in the abstract spider are added to the 
	 *           concrete spider.
	 */
	/*
	 * test:
	 *   s = ("",[(a,)])
	 *     cs = ("y",[(a,)])
	 *     cs = ("",[])
	 *     cs = ("",[(a,)])
	 *     cs = ("",[(a,),(,a)])
	 */
	
	private boolean match(Spider s, ConcreteSpider cs, ConcreteUnitaryDiagram cud) {
		if (!s.getLabel().equals(cs.getLabel()))
		  return false;
		Collection abstractHabitat = (Collection) s.getHabitat().clone();
		Set concreteHabitat = cs.getHabitat(cud);
		Iterator cit = concreteHabitat.iterator();
		while(cit.hasNext()){
			ConcreteZone csf = (ConcreteZone) cit.next();
			if (!abstractHabitat.contains(csf.getAbstractZone()))
				return false;
			else
				abstractHabitat.remove(csf.getAbstractZone());
		}
		if(abstractHabitat.isEmpty())
		  return true;
		// concrete spider is proper subset
		Iterator ait = abstractHabitat.iterator();
		while (ait.hasNext()) {
			Zone z = (Zone) ait.next();
			Set czones = cud.getConcreteZones();
			Iterator czit = czones.iterator();
			while (czit.hasNext()) {
				ConcreteZone cz = (ConcreteZone)czit.next();
				if (cz.getAbstractZone()==z){
					cs.addExistentialFoot(cud,cz);			
				}
			}
		}
		return true;
	}
	
	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {

    	hd("Create Reconciler");
    	Reconciler r = new Reconciler();
    	pr("r="+r);
    	pg();
    	
    	/*
    	{
	    	hd("Tests of intersectContours");
	    	HashSet s1 = new HashSet();
	    	HashSet s2 = new HashSet();
	    	pr("s1="+s1);
	    	pr("s2="+s2);
	    	pr("r.intersectContours(s1,s2)="+r.intersectContours(s1,s2));
	    	pg();
	    
	    	s2.add(Contour.getContour("b"));
	    	pr("s1="+s1);
	    	pr("s2="+s2);
	    	pr("r.intersectContours(s1,s2)="+r.intersectContours(s1,s2));
	    	pg();
	    	
	    	s2.add(Contour.getContour("c"));
	    	s2.add(Contour.getContour("d"));
	
	    	java.awt.Polygon p1 = new java.awt.Polygon();
	    	p1.addPoint(1,2);
	    	s1.add(new ConcreteContour(Contour.getContour("a"), p1));
	    	
	    	java.awt.Polygon p2 = new java.awt.Polygon();
	    	p2.addPoint(3,4);    	
	    	s1.add(new ConcreteContour(Contour.getContour("b"), p2));
	    	
	    	java.awt.Polygon p3 = new java.awt.Polygon();
	    	p3.addPoint(5,6);   	
	    	s1.add(new ConcreteContour(Contour.getContour("c"), p3));
	    	
	    	pr("s1="+s1);
	    	pr("s2="+s2);
	    	pr("r.intersectContours(s1,s2)="+r.intersectContours(s1,s2));
	    	pg();
    	}
    	
    	
    	{

    		hd("set up concrete diagram");
    		DecorationBox d = new DecorationBox("d","",new java.awt.Polygon());
    		ConcreteUnitaryDiagram cud = new ConcreteUnitaryDiagram(d);
       		pr("cud.getConcreteZoneSet().getZones()="+
        		cud.getConcreteZoneSet().getZones());
       		pg();
        	ConcreteZoneSet czs = new ConcreteZoneSet();
    		
	    	java.awt.Polygon p1 = new java.awt.Polygon();
	    	p1.addPoint(1,2);
	    	ConcreteContour ca = 
	    		new ConcreteContour(Contour.getContour("a"), p1);
	    	pr("ca="+ca);
	    	pr("add ca to cud");
	    	cud.addConcreteContour(ca);
       		pr("cud.getConcreteZoneSet().getZones()="+
            	cud.getConcreteZoneSet().getZones());
           	pg();
	    	
	    	pr("shade zone (ad,)");
	    	Collection czones = cud.getConcreteZoneSet().getZones();
	    	Iterator zit = czones.iterator();
	    	boolean hasa;
	    	boolean hasd;
	    	
	    	while (zit.hasNext()){
	    		ConcreteZone z = (ConcreteZone) zit.next();
	    		Collection ccontours = z.getContainingPolygonals();
	    		Iterator cit = ccontours.iterator();
	    		hasa = false;
	    		hasd = false;
	    		while (cit.hasNext()) {
	    			DiagramItem cc = (DiagramItem) cit.next();
	    			if (cc.getLabel().equals("a")) hasa = true;
	    			if (cc.getLabel().equals("d")) hasd = true;
	    		}
	    		if (hasa && hasd) z.setBackColor(Color.GRAY);
	    	}
    		pr("cud="+cud);
    		pr("cud.getConcreteZoneSet().getZones()="+
    			cud.getConcreteZoneSet().getZones());
    		pg();
    		
    		hd("set up abstract diagram");
			java.util.Vector va = new java.util.Vector();
	    	va.add(Contour.getContour("a"));
			java.util.Vector v = new java.util.Vector();
	    	
			Zone az = Zone.getZone(va,v);
			pr("az="+az);
			Zone za = Zone.getZone(v,va);
			pr("za="+za);
			pg();
			pr("add az and za to ud");
			java.util.Vector zones  = new java.util.Vector();
			zones.add(az);
			zones.add(za);
			UnitaryDiagram ud = new UnitaryDiagram(zones,"ad");
			pr("ud="+ud);
    		pr("ud.getCloneOfContourSet()="+ud.getCloneOfContourSet());
    		pr("ud.getCloneOfZoneSet()="+ud.getCloneOfZoneSet());
    		pr("ud.getCloneOfShadedZones()="+ud.getCloneOfShadedZones());
			pg();
			
			pr("shade za");
			ud.shadeZone(za);
    		pr("ud="+ud);
    		pr("ud.getCloneOfContourSet()="+ud.getCloneOfContourSet());
    		pr("ud.getCloneOfZoneSet()="+ud.getCloneOfZoneSet());
    		pr("ud.getCloneOfShadedZones()="+ud.getCloneOfShadedZones());
    		pg();
    		
    		pr("reconcile cud and ud giving newcud");
    		ConcreteUnitaryDiagram newcud = r.reconcile(cud,ud);
    		pr("newcud="+newcud);
    		pr("newcud.getConcreteZoneSet().getZones()="+
    		    newcud.getConcreteZoneSet().getZones());
    		pg();
    	}
    	*/
    	/*
    	{
    		hd("Test of intersectSpiders");    		
    		HashSet cditems = new HashSet();
    		ArrayList aditems = new ArrayList();
    		
    		// build a concrete diagram with contours a and b (a includes b)
    		Polygon pd = new Polygon();
    		pd.addPoint(-5,5);
    		pd.addPoint(5,5);
    		pd.addPoint(5,-5);
    		pd.addPoint(-5,-5);
    		DecorationBox d = new DecorationBox("","",pd);
    		
    		ConcreteUnitaryDiagram cud = new ConcreteUnitaryDiagram(d);
    		pr("aditems="+aditems);
    		pr("cditems="+cditems);
    		pr("r.intersectSpiders(cditems,aditems,cud)="+
    		    r.intersectSpiders(cditems,aditems,cud));
    		pg();
    		
    		// build two contours a and b
    		Contour a = Contour.getContour("a");
    		Contour b = Contour.getContour("b");

    		Polygon pa = new Polygon();
    		pa.addPoint(-3,3);
    		pa.addPoint(3,3);
    		pa.addPoint(3,-3);
    		pa.addPoint(-3,-3);
    		ConcreteContour ca = new ConcreteContour(a,pa);
    		cud.addConcreteContour(ca);
    		
    		Polygon pb = new Polygon();
    		pb.addPoint(-1,1);
    		pb.addPoint(1,1);
    		pb.addPoint(1,-1);
    		pb.addPoint(-1,-1);
    		ConcreteContour cb = new ConcreteContour(b,pb);
    		cud.addConcreteContour(cb);
   		
    		// build three zones x,y and z (abstract and concrete)
    		// x is outside a and b
    		// y is inside a, outside b
    		// z is inside a and b
    		Vector v = new Vector();
    		Vector vab = new Vector();
    		vab.add(a);
    		vab.add(b);
    		Zone x = Zone.getZone(v,vab);
    		
    		Vector va = new Vector();
    		va.add(a);
    		Vector vb = new Vector();
    		vb.add(b);
    		Zone y = Zone.getZone(va,vb);
    		Zone z = Zone.getZone(vab,v);
    		
    		Vector cvd = new Vector();
    		cvd.add(d);
    		Vector cvab = new Vector();
    		cvab.add(ca);
    		cvab.add(cb);
    		ConcreteZone cx = new ConcreteZone(cvd,cvab);
    		
    		Vector cvda = new Vector();
    		cvda.add(d);
    		cvda.add(ca);
    		Vector cvb = new Vector();
    		cvb.add(cb);
    		Vector cv = new Vector();
    		ConcreteZone cy = new ConcreteZone(cvda,cvb);
       		Vector cvdab = new Vector();
       		cvdab.add(d);
    		cvdab.add(ca);
    		cvdab.add(cb);   		
    		ConcreteZone cz = new ConcreteZone(cvdab,cv);
    		
    		// build abstract spider s1(x,y) and add to aditems
    		Vector xy = new Vector();
    		xy.add(x);
    		xy.add(y);
    		Spider s1 = Spider.getAbstractSpider("s1",xy);
    		aditems.add(s1);
    		pr("aditems="+aditems);
    		pr("cditems="+cditems);
    		pr("r.intersectSpiders(cditems,aditems,cud)="+
        		    r.intersectSpiders(cditems,aditems,cud));
       		pg();
        		
    		// build abstract spider s2(y,z) and add to aditems
    		Vector yz = new Vector();
    		yz.add(y);
    		yz.add(z);
    		Spider s2 = Spider.getAbstractSpider("s2",yz);
    		aditems.add(s2);
    		
    		// build concrete spider s2(y2,z1) and s3(z2,x1) 
    		// and add to cditems and cud   		
    		ConcreteSpider cs2 = new ConcreteSpider();
    		ConcreteSpiderFoot fy2 = 
    			new ConcreteSpiderFoot("s2",ConcreteSpiderFoot.NORMAL);
    		fy2.setLocation(0,-2);
    		ConcreteSpiderFoot fz1 = 
    			new ConcreteSpiderFoot("",ConcreteSpiderFoot.NORMAL);
    		fz1.setLocation(0,0);
    		ConcreteSpiderLeg lyz = 
    			new ConcreteSpiderLeg("lyz",fy2,fz1,ConcreteSpiderLeg.NORMAL);
    		cs2.addFoot(fy2);
    		cs2.addFoot(fz1);
    		cs2.addLeg(lyz);
    		cditems.add(cs2);
    		cud.addConcreteFoot(fy2);
    		cud.addConcreteFoot(fz1);
    		cud.addConcreteLeg(lyz);
    		
    		ConcreteSpider cs3 = new ConcreteSpider();
    		ConcreteSpiderFoot fz2 = 
    			new ConcreteSpiderFoot("s3",ConcreteSpiderFoot.NORMAL);
    		fz2.setLocation(0,0);
    		ConcreteSpiderFoot fx1 = 
    			new ConcreteSpiderFoot("",ConcreteSpiderFoot.NORMAL);
    		fx1.setLocation(4,0);
    		ConcreteSpiderLeg lzx = 
    			new ConcreteSpiderLeg("lzx",fz2,fx1,ConcreteSpiderLeg.NORMAL);
    		cs3.addFoot(fz2);
    		cs3.addFoot(fx1);
    		cs3.addLeg(lzx);
    		cditems.add(cs3);
    		cud.addConcreteFoot(fz2);
    		cud.addConcreteFoot(fx1);
    		cud.addConcreteLeg(lzx);
    		pr("aditems="+aditems);
    		pr("cditems="+cditems);
    		pr("r.intersectSpiders(cditems,aditems,cud)="+
        		    r.intersectSpiders(cditems,aditems,cud));
       		pg();    		
    		
    	}
    	*/
    	/*
    	{
    		hd("Test of intersectArrows");    		
    		HashSet cditems = new HashSet();
    		ArrayList aditems = new ArrayList();
    		
    		// build a concrete diagram with contours a and b (a includes b)
    		Polygon pd = new Polygon();
    		pd.addPoint(-5,5);
    		pd.addPoint(5,5);
    		pd.addPoint(5,-5);
    		pd.addPoint(-5,-5);
    		DecorationBox d = new DecorationBox("d","",pd);
    		
    		ConcreteUnitaryDiagram cud = new ConcreteUnitaryDiagram(d);
    		pr("aditems="+aditems);
    		pr("cditems="+cditems);
    		pr("r.intersectArrows(cditems,aditems,cud)="+
    		    r.intersectArrows(cditems,aditems,cud));
    		pg();
    		
    		// build two contours a and b
    		Contour a = Contour.getContour("a");
    		Contour b = Contour.getContour("b");

    		Polygon pa = new Polygon();
    		pa.addPoint(-3,3);
    		pa.addPoint(3,3);
    		pa.addPoint(3,-3);
    		pa.addPoint(-3,-3);
    		ConcreteContour ca = new ConcreteContour(a,pa);
    		cud.addConcreteContour(ca);
    		
    		Polygon pb = new Polygon();
    		pb.addPoint(-1,1);
    		pb.addPoint(1,1);
    		pb.addPoint(1,-1);
    		pb.addPoint(-1,-1);
    		ConcreteContour cb = new ConcreteContour(b,pb);
    		cud.addConcreteContour(cb);
    		
    		// build abstract arrow ar1 from a to a
    		Arrow ar1 = new Arrow(a,a);
    		aditems.add(ar1);
    		pr("aditems="+aditems);
    		pr("cditems="+cditems);
    		pr("r.intersectArrows(cditems,aditems,cud)="+
        		    r.intersectArrows(cditems,aditems,cud));
       		pg();
        		
    		
    		// build abstract and concrete arrow ar2 from a to b
    		Arrow ar2 = new Arrow(a,b);
    		aditems.add(ar2);

    		ConcreteArc car2 = new ConcreteArc(ca,cb);
    		car2.setLabel("car2");
    		cditems.add(car2);
  		
    		// build concrete arrow ar3 from b to a
    		ConcreteArc car3 = new ConcreteArc(cb,ca);
    		cditems.add(car3);
    		car3.setLabel("car3");
    		pr("aditems="+aditems);
    		pr("cditems="+cditems);
    		pr("r.intersectArrows(cditems,aditems,cud)="+
        		    r.intersectArrows(cditems,aditems,cud));
       		pg();    		
    		
    	}
    	*/
    	/*
    	{
    		// tests of match
 
    		// build abstract diagram
    		Contour a = Contour.getContour("a");
    		Vector va = new Vector();
    		va.add(a);
    		Vector v = new Vector();
			Zone az = Zone.getZone(va,v);
			pr("az="+az);
			Zone za = Zone.getZone(v,va);
			pr("za="+za);
			pg();
			
			Vector hx = new Vector();
			hx.add(az);
    		Spider s = new Spider("",hx);
    		pr("s="+s);
    		pg();
    		
    		//build concrete diagram with contour a
    		Polygon pd = new Polygon();
    		pd.addPoint(-5,5);
    		pd.addPoint(5,5);
    		pd.addPoint(5,-5);
    		pd.addPoint(-5,-5);
    		DecorationBox d = new DecorationBox("","",pd);
    		
    		ConcreteUnitaryDiagram cud = new ConcreteUnitaryDiagram(d);
    		
    		// build contour a 
    		Polygon pa = new Polygon();
    		pa.addPoint(-3,3);
    		pa.addPoint(3,3);
    		pa.addPoint(3,-3);
    		pa.addPoint(-3,-3);
    		ConcreteContour ca = new ConcreteContour(a,pa);
    		cud.addConcreteContour(ca);
    		// build concrete zones cdaz, cdza		
    		
    		Vector cvd = new Vector();
    		cvd.add(d);
    		Vector cva = new Vector();
    		cva.add(ca);
    		ConcreteZone cdza = new ConcreteZone(cvd,cva);
    		
    		Vector cvda = new Vector();
    		cvda.add(d);
    		cvda.add(ca);
    		Vector cv = new Vector();
    		ConcreteZone cdaz = new ConcreteZone(cvda,cv);
    		
    		// build concrete spider ("y",[(a,)])
    		
    		ConcreteSpider cs1 = new ConcreteSpider();
    		ConcreteSpiderFoot fcs1 = 
    			new ConcreteSpiderFoot("y",ConcreteSpiderFoot.NORMAL);
    		fcs1.setLocation(0,0);
    		cs1.addFoot(fcs1);
    		pr("cs1="+cs1);
    		pg();
    		
    		pr("match(s,cs1,cud)(should be false)="+r.match(s,cs1,cud));
    		pr("cs1="+cs1);
    		pg();
    		
    		ConcreteSpider cs2 = new ConcreteSpider();
    		pr("cs2="+cs2);
    		pr("match(s,cs2,cud)(should be true and foot added)="+r.match(s,cs2,cud));
    		pr("cs2="+cs2);
    		pg();
   		
    		ConcreteSpider cs3 = new ConcreteSpider();
    		ConcreteSpiderFoot fcs3 = 
    			new ConcreteSpiderFoot("",ConcreteSpiderFoot.NORMAL);
    		fcs3.setLocation(0,0);
    		cs3.addFoot(fcs3);
    		pr("cs3="+cs3);
    		pr("match(s,cs3,cud)(should be true)="+r.match(s,cs3,cud));
    		pr("cs3="+cs3);
    		pg();
   		
    		ConcreteSpider cs4 = new ConcreteSpider();
    		ConcreteSpiderFoot fcs4 = 
    			new ConcreteSpiderFoot("",ConcreteSpiderFoot.NORMAL);
    		fcs4.setLocation(4,0);
    		cs4.addFoot(fcs3);
    		cs4.addFoot(fcs4);
    		pr("cs4="+cs4);
    		pr("match(s,cs4,cud)(should be false)="+r.match(s,cs4,cud));
    		pr("cs4="+cs4);
    		pg();
   		
    	}
    	*/
	/*
    }
    */	
}
