package rwd.rules.nospec;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.Diagram;
import rwd.abstractsyntax.spiderdiagram.FalseDiagram;
import rwd.abstractsyntax.spiderdiagram.NotCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.OrCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.rules.Rule;
import rwd.rules.application.RuleAppliesToJustADiagram;
import rwd.rules.application.RuleApplicationSpec;
import java.util.Vector;
import java.util.Iterator;

public class ReverseNegationEquivalenceRule extends RuleNoSpec {
    private ReverseNegationEquivalenceRule() {
        super("reverse negation equivalence");
    }
    private static Rule example = null;

    public static Rule getRule() {
        if (example == null) {
            example = new ReverseNegationEquivalenceRule();
        }
        return example;
    }
    public boolean isInvertible() {
        return true;
    }

    public boolean canApplyTo(Diagram d) {
        // the input is a disjunction of unitary alpha diagrams di
        // each in Venn form with the same contours
        // each has exactly one non-blank zone z
        // either
        //    z is shaded in all d1..dn and containing 1,2,3...n spiders
        // or 
        //    z is shaded in all d1..d(n-1) containing 1,2,..,n-1 spiders
        //    and dn contains no shading and n spiders in z

        if (!(d instanceof OrCompoundDiagram)) {
            return false;
        }
        Vector children = ((OrCompoundDiagram) d).getChildren();
        for (int i = 0; i < children.size(); i++) {
            Diagram child = (Diagram) children.get(i);
            if (!(child instanceof UnitaryDiagram)) {
                return false;
            }
            if (child instanceof FalseDiagram) {
                return false;
            }
            Vector spiders = ((UnitaryDiagram) child).getSpiders();
            for (int j = 0; j < spiders.size(); j++) {
                Spider s = (Spider) spiders.get(j);
                if (s.getHabitat().size() != 1) {
                    return false;
                }
            }
        }
        UnitaryDiagram d1 = (UnitaryDiagram) children.get(0);
        // check contour sets are equal
        Vector contours = d1.getCloneOfContourSet();
        for (int childNum = 1; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            Vector otherContours = child.getCloneOfContourSet();
            if (!contours.containsAll(otherContours)
                || !otherContours.containsAll(contours)) {
                return false;
            }
        }
        // check all are in Venn form
        Vector zones = UnitaryDiagram.allPossibleZones(contours);
        for (int childNum = 0; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            Vector otherZones = child.getCloneOfZoneSet();
            if (!zones.containsAll(otherZones)
                || !otherZones.containsAll(zones)) {
                return false;
            }
        }
        // find one zone in d1 which is populated
        Zone populatedZone = null;
        Vector spiders = ((UnitaryDiagram) d1).getSpiders();
        Iterator zi = d1.getZoneIterator();
        ziLoop : while (zi.hasNext()) {
            Zone z = (Zone) zi.next();
            if (d1.containsShaded(z)) {
                if (populatedZone != null) {
                    return false;
                    //System.out.println("two zones populated");
                } else {
                    populatedZone = z;
                    continue ziLoop;
                }
            } else {
                // look for spiders in z
                for (int j = 0; j < spiders.size(); j++) {
                    Spider s = (Spider) spiders.get(j);
                    if (s.getHabitat().contains(z)) {
                        if (populatedZone != null) {
                            //System.out.println("two zones populated");
                            return false;
                        } else {
                            populatedZone = z;
                            continue ziLoop;
                        }
                    }
                }
            }
        }
        if (populatedZone == null) {
            //System.out.println("no zone populated");
            return false;
        }
        Vector justTheZone = new Vector();
        justTheZone.add(populatedZone);
        for (int childNum = 1; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            if (!(justTheZone.containsAll(child.getCloneOfShadedZones()))) {
                //System.out.println("shading outside populated zone");
                return false;
            }
            Vector spidersInOtherChild = child.getSpiders();
            for (int j = 0; j < spidersInOtherChild.size(); j++) {
                Spider s = (Spider) spidersInOtherChild.get(j);
                if (!justTheZone.containsAll(s.getHabitat())) {
                    //System.out.println("spider outside populated zone");
                    return false;
                }
            }
        }
        // make sure that we have d1, d2, ... d(n-1) which are
        // all shaded in z and have 1, 2, .. n-1 spiders in z
        for (int i = 0; i < children.size() - 1; i++) {
            // build di then scan for its presence
            UnitaryDiagram di = new UnitaryDiagram(zones, "ud");
            di.shadeZone(populatedZone);
            for (int j = 0; j < i; j++) {
                di.addSpider((Vector) justTheZone.clone());
            }
            boolean foundMatch = false;
            for (int childNum = 0; childNum < children.size(); childNum++) {
                // check that populatedzone is only thing populated here
                UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
                if (child.equals(di)) {
                    foundMatch = true;
                }
            }
            if (!foundMatch) {
                //System.out.println("didn't find match for "+i+"th intermediate");
                return false;
            }
        }
        boolean foundMatch = false;
        // build two possible dn's
        UnitaryDiagram dn = new UnitaryDiagram(zones, "ud");
        for (int j = 1; j < children.size(); j++) {
            dn.addSpider((Vector) justTheZone.clone());
        }
        for (int childNum = 0; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            if (child.equals(dn)) {
                foundMatch = true;
            }
        }
        dn.shadeZone(populatedZone);
        dn.removeSpider((Spider)dn.getSpiders().get(0));
        for (int childNum = 0; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            if (child.equals(dn)) {
                foundMatch = true;
            }
        }        
        if(!foundMatch){
            //System.out.println("didn't find match for last");
        }
        return foundMatch;
    }
    public Diagram applyRule(RuleApplicationSpec r) {
        RuleAppliesToJustADiagram rajd = (RuleAppliesToJustADiagram) r;
        Diagram d = rajd.d;
        //System.out.println("applying rwd.rules to "+d);
        // the input is a disjunction of unitary alpha diagrams di
        // each in Venn form with the same contours
        // each has exactly one non-blank zone z
        // either
        //    z is shaded in all d1..dn and containing 1,2,3...n spiders
        // or 
        //    z is shaded in all d1..d(n-1) containing 1,2,..,n-1 spiders
        //    and dn contains no shading and n spiders in z

        if (!(d instanceof OrCompoundDiagram)) {
            throw new IllegalArgumentException();
        }
        Vector children = ((OrCompoundDiagram) d).getChildren();
        for (int i = 0; i < children.size(); i++) {
            Diagram child = (Diagram) children.get(i);
            if (!(child instanceof UnitaryDiagram)) {
                throw new IllegalArgumentException();
            }
            if (child instanceof FalseDiagram) {
                throw new IllegalArgumentException();
            }
            Vector spiders = ((UnitaryDiagram) child).getSpiders();
            for (int j = 0; j < spiders.size(); j++) {
                Spider s = (Spider) spiders.get(j);
                if (s.getHabitat().size() != 1) {
                    throw new IllegalArgumentException();
                }
            }
        }
        UnitaryDiagram d1 = (UnitaryDiagram) children.get(0);
        // check contour sets are equal
        Vector contours = d1.getCloneOfContourSet();
        for (int childNum = 1; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            Vector otherContours = child.getCloneOfContourSet();
            if (!contours.containsAll(otherContours)
                || !otherContours.containsAll(contours)) {
                    throw new IllegalArgumentException();
            }
        }
        // check all are in Venn form
        Vector zones = UnitaryDiagram.allPossibleZones(contours);
        for (int childNum = 0; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            Vector otherZones = child.getCloneOfZoneSet();
            if (!zones.containsAll(otherZones)
                || !otherZones.containsAll(zones)) {
                    throw new IllegalArgumentException();
            }
        }
        // find one zone in d1 which is populated
        Zone populatedZone = null;
        Vector spiders = ((UnitaryDiagram) d1).getSpiders();
        Iterator zi = d1.getZoneIterator();
        ziLoop : while (zi.hasNext()) {
            Zone z = (Zone) zi.next();
            if (d1.containsShaded(z)) {
                if (populatedZone != null) {
                    throw new IllegalArgumentException();
                    //System.out.println("two zones populated");
                } else {
                    populatedZone = z;
                    continue ziLoop;
                }
            } else {
                // look for spiders in z
                for (int j = 0; j < spiders.size(); j++) {
                    Spider s = (Spider) spiders.get(j);
                    if (s.getHabitat().contains(z)) {
                        if (populatedZone != null) {
                            //System.out.println("two zones populated");
                            throw new IllegalArgumentException();
                        } else {
                            populatedZone = z;
                            continue ziLoop;
                        }
                    }
                }
            }
        }
        if (populatedZone == null) {
            //System.out.println("no zone populated");
            throw new IllegalArgumentException();
        }
        Vector justTheZone = new Vector();
        justTheZone.add(populatedZone);
        for (int childNum = 1; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            if (!(justTheZone.containsAll(child.getCloneOfShadedZones()))) {
                //System.out.println("shading outside populated zone");
                throw new IllegalArgumentException();
            }
            Vector spidersInOtherChild = child.getSpiders();
            for (int j = 0; j < spidersInOtherChild.size(); j++) {
                Spider s = (Spider) spidersInOtherChild.get(j);
                if (!justTheZone.containsAll(s.getHabitat())) {
                    //System.out.println("spider outside populated zone");
                    throw new IllegalArgumentException();
                }
            }
        }
        // make sure that we have d1, d2, ... d(n-1) which are
        // all shaded in z and have 1, 2, .. n-1 spiders in z
        for (int i = 0; i < children.size() - 1; i++) {
            // build di then scan for its presence
            UnitaryDiagram di = new UnitaryDiagram(zones, "ud");
            di.shadeZone(populatedZone);
            for (int j = 0; j < i; j++) {
                di.addSpider((Vector) justTheZone.clone());
            }
            boolean foundMatch = false;
            for (int childNum = 0; childNum < children.size(); childNum++) {
                // check that populatedzone is only thing populated here
                UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
                if (child.equals(di)) {
                    foundMatch = true;
                }
            }
            if (!foundMatch) {
                //System.out.println("didn't find match for "+i+"th intermediate");
                throw new IllegalArgumentException();
            }
        }
        boolean foundMatch = false;
        // build two possible dn's
        UnitaryDiagram dn = new UnitaryDiagram(zones, "ud");
        for (int j = 1; j < children.size(); j++) {
            dn.addSpider((Vector) justTheZone.clone());
        }
        for (int childNum = 0; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            if (child.equals(dn)) {
                dn.removeSpider((Spider)dn.getSpiders().get(0));
                dn.shadeZone(populatedZone);
                //System.out.println("num shaded zones is 1="+dn.getNumShadedZones());
                Vector singleton = new Vector();
                singleton.add(dn);
                return new NotCompoundDiagram(singleton,"not ud");
            }
        }
        dn.shadeZone(populatedZone);
        dn.removeSpider((Spider)dn.getSpiders().get(0));
        for (int childNum = 0; childNum < children.size(); childNum++) {
            // check that populatedzone is only thing populated here
            UnitaryDiagram child = (UnitaryDiagram) children.get(childNum);
            if (child.equals(dn)) {
                dn.addSpider((Vector) justTheZone.clone());
                dn.unshade(populatedZone);
                Vector singleton = new Vector();
                singleton.add(dn);
                return new NotCompoundDiagram(singleton,"not ud");
            }
        }        
        throw new IllegalArgumentException();
    }
}
