package rwd.views.abstracttoconcrete.forcemodel.layout;

import rwd.views.abstracttoconcrete.forcemodel.*;

/**
 * @author  $Author: op $
 * @version $Id: ForceLayout.java,v 1.5 2004/10/20 10:08:39 op Exp $
 */
public class ForceLayout implements Layout {
    
    public ForceLayout() {
        this(1000, 2, 0.01, 100);
    }
    
    /**
     * Constructs a force directed layout object with the specified
     * parameters.
     */
    public ForceLayout(long iterations, double k, double c, double maxRepulsiveForceDistance) {
        _iterations = iterations;
        _k = k;
        _c = c;
        _maxRepulsiveForceDistance = maxRepulsiveForceDistance;
    }
    
    public void apply(Graph graph) {
        
        Node[] nodes = graph.getNodeArray();
        Edge[] edges = graph.getEdgeArray();
        
        // For each iteration...
        for (int it = 0; it < _iterations; it++) {
            
            // Calculate forces acting on nodes due to node-node repulsions...
            for (int a = 0; a < nodes.length; a++) {
                for (int b = a + 1; b < nodes.length; b++) {
                    Node nodeA = nodes[a];
                    Node nodeB = nodes[b];
                    
                    double deltaX = nodeB.getX() - nodeA.getX();
                    double deltaY = nodeB.getY() - nodeA.getY();
                    
                    double distanceSquared = deltaX * deltaX + deltaY * deltaY;
                    
                    if (distanceSquared < 0.01) {
                        deltaX = Math.random() / 10 + 0.1;
                        deltaY = Math.random() / 10 + 0.1;
                        distanceSquared = deltaX * deltaX + deltaY * deltaY;
                    }
                    
                    double distance = Math.sqrt(distanceSquared);
                    
                    if (distance < _maxRepulsiveForceDistance) {
                        double repulsiveForce = (_k * _k / distance);
                        
                        nodeB.setFX(nodeB.getFX() + (repulsiveForce * deltaX / distance));
                        nodeB.setFY(nodeB.getFY() + (repulsiveForce * deltaY / distance));
                        nodeA.setFX(nodeA.getFX() - (repulsiveForce * deltaX / distance));
                        nodeA.setFY(nodeA.getFY() - (repulsiveForce * deltaY / distance));
                    }
                }
            }
            
            // Calculate forces acting on nodes due to edge attractions
            for (int e = 0; e < edges.length; e++) {
                Edge edge = edges[e];
                Node nodeA = edge.getSource();
                Node nodeB = edge.getTarget();
                
                double deltaX = nodeB.getX() - nodeA.getX();
                double deltaY = nodeB.getY() - nodeA.getY();
                
                double distanceSquared = deltaX * deltaX + deltaY * deltaY;
                
                // Avoid division by zero error or Nodes flying off to
                // infinity.  Pretend there is an arbitrary distance between
                // the Nodes.
                if (distanceSquared < 0.01) {
                    deltaX = Math.random() / 10 + 0.1;
                    deltaY = Math.random() / 10 + 0.1;
                    distanceSquared = deltaX * deltaX + deltaY * deltaY;
                }
                
                double distance = Math.sqrt(distanceSquared);
                
                if (distance >  _maxRepulsiveForceDistance) {
                    distance = _maxRepulsiveForceDistance;
                }
                
                distanceSquared = distance * distance;
                
                double attractiveForce = (distanceSquared - _k * _k) / _k;
                
                //if (edge.isSamePerson()) {
                //    attractiveForce *= 4;
                //}
            
                nodeB.setFX(nodeB.getFX() - attractiveForce * deltaX / distance);
                nodeB.setFY(nodeB.getFY() - attractiveForce * deltaY / distance);
                nodeA.setFX(nodeA.getFX() + attractiveForce * deltaX / distance);
                nodeA.setFY(nodeA.getFY() + attractiveForce * deltaY / distance);
                
            }
            
            // Now move each node to its new location...
            
            for (int a = 0; a < nodes.length; a++) {
                Node node = nodes[a];
                
                double xMovement = _c * node.getFX();
                double yMovement = _c * node.getFY();
                
                // Limit movement values to stop nodes flying into oblivion.
                if (xMovement > 50) {
                    xMovement = 50;
                }
                if (yMovement > 50) {
                    yMovement = 50;
                }
                
                node.setX(node.getX() + xMovement);
                node.setY(node.getY() + yMovement);
                
                // Reset the forces
                node.setFX(0);
                node.setFY(0);
            }
            
        }
        
    }

    
    private long _iterations;
    private double _k;
    private double _c;
    private double _maxRepulsiveForceDistance;
    
}
