package rwd.views.graph;
import java.util.TreeSet;
import java.util.Vector;
import java.util.Iterator;
import java.util.NoSuchElementException;
/**
 * A directed edge is an edge with orientation.
 * The two nodes are stored as "to" and "from".
 */
public class DirectedEdge implements Comparable {
	private Node from;
	private Node to;
	private Comparable label;
	private double a, b, c, d;
	/////////////////////////////////////////////////////////////////////////
	/**
	 * Defines a directed edge from and to the specified Nodes.
	 * If the nodes correspond to equal AbNodes then a
	 * SelfLoop exception is thrown.
	 */
	public DirectedEdge(
		Node n1,
		Node n2) { //System.out.println("building de from "+n1 +" to "+n2);
		from = n1;
		to = n2;
		a = n1.getPosition().getX();
		b = n1.getPosition().getY();
		c = n2.getPosition().getX();
		d = n2.getPosition().getY();
		try {
			label = (Comparable) (n1.symmDiff(n2).iterator().next());
		} catch (NoSuchElementException x) {
		}
	}
	///////////////////////////////////////////////////////////////////////////
	public Node from() {
		return from;
	}
	///////////////////////////////////////////////////////////////////////////
	public Node to() {
		return to;
	}
	///////////////////////////////////////////////////////////////////////////
	public int compareTo(Object o) {
		DirectedEdge de = (DirectedEdge) o;
		int fromCompare = from().compareTo(de.from());
		if (fromCompare != 0) {
			return fromCompare;
		} else {
			return to().compareTo(de.to());
		}
	}
	///////////////////////////////////////////////////////////////////////////
	public String toString() {
		return "" + from + "->" + to;
	}
	///////////////////////////////////////////////////////////////////////////
	/**
	 * Returns the angle in radians anticlockwise from a horizontal.
	 * The angle lies between 0 and (strictly less than) 2Pi.
	 */
	public double angleFromZero() { //System.out.println("finding angle of "+this);
		double xDisplacement =
			to().getPosition().getX() - from().getPosition().getX();
		double yDisplacement =
			to().getPosition().getY() - from().getPosition().getY();
		if (xDisplacement == 0 && yDisplacement > 0) {
			return Math.PI / 2;
		}
		if (xDisplacement == 0 && yDisplacement < 0) {
			return 3 * Math.PI / 2;
		}
		if (xDisplacement > 0 && yDisplacement == 0) {
			return 0;
		}
		if (xDisplacement < 0 && yDisplacement == 0) {
			return Math.PI;
		}
		double tanGuess = Math.atan(yDisplacement / xDisplacement);
		if (yDisplacement < 0 && xDisplacement > 0) {
			tanGuess = tanGuess + 2 * Math.PI;
		}
		if (xDisplacement < 0) {
			tanGuess = tanGuess + Math.PI;
		}
		return tanGuess;
	}
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Returns an angle in radians from this to the paramater DirectedEdge.
	 * The angle is a measure of how sharply the corner "turns left".
	 * The result is always (strictly) more than 0 and (weakly)
	 * less than 2Pi.
	 */
	public double angleTo(DirectedEdge de) {
		double gapAngle = (this.reverse()).angleFromZero() - de.angleFromZero();
		while (gapAngle <= 0) {
			gapAngle = gapAngle + 2 * Math.PI;
		}
		while (gapAngle > 2 * Math.PI) {
			gapAngle = gapAngle - 2 * Math.PI;
		}
		return gapAngle;
	}
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Returns an angle in radians from this to the paramater DirectedEdge.
	 * The angle is a measure of how sharply the corner "turns left".
	 * The result is always (weakly) more than 0 and (strictly)
	 * less than 2Pi.
	 */
	public double smallAngleTo(DirectedEdge de) {
		double angle = angleTo(de);
		if (angle >= 2 * Math.PI - 0.0001) {
			angle = angle - 2 * Math.PI;
		}
		return angle;
	}
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Returns the directed edge in given context which gives
	 * the sharpest turn left.
	 * Imagine going along this edge and meeting all edges in
	 * context, starting at the end point.  Take the "first left".
	 * Returns null if the context is empty.
	 */
	public DirectedEdge turnLeft(TreeSet context) {
		TreeSet contextCopy = (TreeSet) (context.clone());
		TreeSet contextCopyForIteration = (TreeSet) (context.clone());
		Iterator i = contextCopyForIteration.iterator();
		while (i.hasNext()) {
			DirectedEdge de = (DirectedEdge) (i.next());
			if (de.from().compareTo(to()) != 0) {
				//System.out.println(""+de.from()+" is not "+to()+" so chuck out "+de);
				contextCopy.remove(de);
			}
		}
		//System.out.println("leaves only "+context);
		double bestAngle = 3 * Math.PI;
		DirectedEdge bestDEdge = null;
		i = contextCopy.iterator();
		while (i.hasNext()) {
			DirectedEdge de = (DirectedEdge) (i.next());
			double thisAngle = angleTo(de);
			if (thisAngle < bestAngle) {
				bestAngle = thisAngle;
				bestDEdge = de;
			}
		}
		return bestDEdge;
	}
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Returns a new DirectedEdge with opposite orientation.
	 */
	public DirectedEdge reverse() {
		return new DirectedEdge(to, from);
	}
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Returns a TreeSet of 2n directed edges for the n edges of given rwd.views.graph
	 */
	public static TreeSet getAllDirectedEdges(Graph g) {
		TreeSet result = new TreeSet();
		Iterator i = g.edges().iterator();
		while (i.hasNext()) {
			Edge e = (Edge) (i.next());
			Iterator j = e.ends().iterator();
			Node n1 = (Node) (j.next());
			Node n2 = (Node) (j.next());
			result.add(new DirectedEdge(n1, n2));
			result.add(new DirectedEdge(n2, n1));
		}
		return result;
	}
	///////////////////////////////////////////////////////////////////////////
	public Comparable label() {
		return label;
	}
	/**
	* Checks if the node is in the internal portion of the edge.
	**/
	public boolean crosses(Node n) {
		double p = n.getPosition().getX();
		double q = n.getPosition().getY();
		return ((b - q) * (a - c) == (b - d) * (a - p) && isBetween(p, a, c))
			|| ((b - q) * (a - c) == (b - d) * (a - p) && isBetween(q, b, d));
	}
	///////////////////////////////////////////////////////////////////////////
	private boolean isBetween(double middle, double lower, double upper) {
		return (middle - lower) * (middle - upper) < 0;
	}
	///////////////////////////////////////////////////////////////////////////
	public Node midNode(String s) {
		TreeSet label = new TreeSet();
		label.add("mid" + s + Edge.getEdge(from(), to()).label());
		Node n = Node.getNode(label,new TreeSet());
		//??????????????????????????????????????????
		Vector ends = new Vector();
		ends.add(to.getPosition());
		ends.add(from.getPosition());
		n.setConstraints(ends);
		n.setPosition(
			new Position(
				(to.getPosition().getX() + from.getPosition().getX()) / 2,
				(to.getPosition().getY() + from.getPosition().getY()) / 2));
		//System.out.println("the midnode of "+this+" is "+n);
		return n;
	}

	public Edge edge() {
		return Edge.getEdge(from, to);
	}
}
