package rwd.views.smoother;

import java.awt.Point;
import java.awt.Polygon;
import java.util.Vector;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;

import rwd.concretesyntax.ConcreteDiagram;
import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.concretesyntax.eulerdiagram.ConcreteContour;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderFoot;
import rwd.gui.editor.viewers.DiagramViewerInterface;

public class DiagramMover {

	protected DiagramViewerInterface viewer;

	protected SwingWorker worker;
	
	protected List contours;
	protected List feet;
	protected HashMap feetPhases = new HashMap();
	protected HashMap contourPhases = new HashMap();

	// 20 move circuit
	//final int[] deltaX = {1,1,1,1,0,0,0,-1,-1,-1,-1,-1,-1,-1,0,0,0,1,1,1};
	//final int[] deltaY = {0,0,-1,-1,-1,-1,-1,-1,-1,0,0,0,1,1,1,1,1,1,1,0};

	// 64 move circuit
	final int[] deltaX = {1,1,1,1,1,1,1,1,1,1,0,1,0,1,0,0,0,0,-1,0,-1,0,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,0,-1,0,-1,0,0,0,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1};
	final int[] deltaY = {0,0,0,-1,0,-1,0,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,0,-1,0,-1,0,0,0,0,0,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,1,0,0,0};

	// size of circle
	final int circleSize = 10;
	// time for a complete circle, in milliseconds
	final int circuitTime = 2000;

    public DiagramMover(DiagramViewerInterface viewer) {
		this.viewer = viewer;

    }

	public void stopMotion() {
		worker = new SwingWorker() {
			public Object construct() {
				viewer.setMotionFlag(false);
				return (new Object());
			}
		};
		worker.start();
	}

	public void startMotion() {
		if(viewer.getMotionFlag() == true) {
			return;
		}
		viewer.setMotionFlag(true);

		ConcreteDiagram cd = viewer.getConcreteDiagram();
		contours = cd.getConcreteContourSet().getContours();

		feet = cd.getFeetAndLegsSet().getFeet();

		// TBD - store old points to replace when stopped, FastHillClimber does this.
		boolean structureOK = false;
		Random random = new Random();

		// set up the polygon phases
		// this is not very clever at all, it keeps trying
		// random phases for all the contours and spiders
		// until a structure preserving combination is found
		while(!structureOK) {
			feetPhases = new HashMap();
			contourPhases = new HashMap();

			// get spider feet phases
			Iterator feetIt1 = feet.iterator();
			while (feetIt1.hasNext()) {
				ConcreteSpiderFoot csf = (ConcreteSpiderFoot)feetIt1.next();
				Integer polyPhase = new Integer(random.nextInt(deltaX.length));
				feetPhases.put(csf,polyPhase);
			}
			
			// get contour phases
			Iterator contourIt1 = contours.iterator();
			while (contourIt1.hasNext()) {
				ConcreteContour cp = (ConcreteContour)contourIt1.next();
				Polygon p = cp.getPolygon();
				Integer polyPhase = new Integer(random.nextInt(deltaX.length));
				contourPhases.put(cp,polyPhase);
			}
			structureOK = checkPhaseStructure(cd);
		}

		worker = new SwingWorker() {
			public Object construct() {

				// get starting feet positions
				HashMap originalFeetPositions = new HashMap();
				Iterator feetIt2 = feet.iterator();
				while (feetIt2.hasNext()) {
					ConcreteSpiderFoot csf = (ConcreteSpiderFoot)feetIt2.next();
					Point location = new Point(csf.getX(),csf.getY());
					originalFeetPositions.put(csf,location);
				}
				// get starting contour positions
				HashMap originalContourPositions = new HashMap();
				Iterator contourIt2 = contours.iterator();
				while (contourIt2.hasNext()) {
					ConcreteContour cp = (ConcreteContour)contourIt2.next();
					Polygon p = cp.getPolygon();
					Vector originalPoints = FastHillClimber.getPolygonPoints(p);
					originalContourPositions.put(p,originalPoints);
				}

				int phaseStage = 0;
				long startOfCircuitTime = System.currentTimeMillis();
				while(viewer.getMotionFlag() == true) {
					// move the spiders
					Iterator feetIt3 = feet.iterator();
					while (feetIt3.hasNext()) {
						ConcreteSpiderFoot csf = (ConcreteSpiderFoot)feetIt3.next();
						Integer footPhase = (Integer)feetPhases.get(csf);
						int deltaInd = (phaseStage+footPhase.intValue())%deltaX.length;
						csf.setLocation(csf.getX()+deltaX[deltaInd]*circleSize,csf.getY()+deltaY[deltaInd]*circleSize);
					}
					// move the contours
					Iterator contourIt3 = contours.iterator();
					while (contourIt3.hasNext()) {
						ConcreteContour cp = (ConcreteContour)contourIt3.next();
						Polygon p = cp.getPolygon();
						Integer polyPhase = (Integer)contourPhases.get(cp);
						int deltaInd = (phaseStage+polyPhase.intValue())%deltaX.length;
						FastHillClimber.movePolygonPoints(p, deltaX[deltaInd]*circleSize, deltaY[deltaInd]*circleSize);
						cp.autoPlaceLabel();
					}
					
					// this bit drops frames or adds a delay to maintain the desired circuit time
					long expectedTime = (phaseStage+1)*(circuitTime/(deltaX.length+1));
					long currentCircuitTime = System.currentTimeMillis()-startOfCircuitTime;
					if(currentCircuitTime<expectedTime) {
						viewer.redrawImage();
					}
					currentCircuitTime = System.currentTimeMillis()-startOfCircuitTime;
					while(currentCircuitTime<expectedTime) {
						currentCircuitTime = System.currentTimeMillis()-startOfCircuitTime;
					}

					phaseStage++;
					if(phaseStage >= deltaX.length) {
						phaseStage = 0;
						startOfCircuitTime = System.currentTimeMillis();
					}
				}

				// reset original feet positions
				Iterator feetIt4 = feet.iterator();
				while (feetIt4.hasNext()) {
					ConcreteSpiderFoot csf = (ConcreteSpiderFoot)feetIt4.next();
					Point originalFootLocation = (Point)originalFeetPositions.get(csf);
					csf.setLocation((int)originalFootLocation.getX(),(int)originalFootLocation.getY());
				}
				// reset original contour positions
				Iterator contourIt4 = contours.iterator();
				while (contourIt4.hasNext()) {
					ConcreteContour cp = (ConcreteContour)contourIt4.next();
					Polygon p = cp.getPolygon();
					Vector originalContourPoints = (Vector)originalContourPositions.get(p);
					FastHillClimber.setPolygonPoints(originalContourPoints,p);
					cp.autoPlaceLabel();
				}
				viewer.redrawImage();

				return (new Object());
			}
		};

		worker.start();

    }

	private boolean checkPhaseStructure(ConcreteDiagram cd) {
		// get all the cuds in the diagram and setup a structure checker for each
		Set unitaryDiags = cd.getConcreteUnitaryDiagrams();
		HashMap structureCheckers = new HashMap();
		Iterator itUD1 = unitaryDiags.iterator();
		while (itUD1.hasNext()) {
			ConcreteUnitaryDiagram cud = (ConcreteUnitaryDiagram)itUD1.next();
			ZonePresenceChecker zpc = new ZonePresenceChecker(cud); 
			SpiderChecker sc = new SpiderChecker(cud); 
			List checkers = new Vector();
			checkers.add(zpc);
			checkers.add(sc);
			structureCheckers.put(cud, checkers);
		}

		// get starting feet positions
		HashMap originalFeetPositions = new HashMap();
		Iterator feetIt1 = feet.iterator();
		while (feetIt1.hasNext()) {
			ConcreteSpiderFoot csf = (ConcreteSpiderFoot)feetIt1.next();
			Point location = new Point(csf.getX(),csf.getY());
			originalFeetPositions.put(csf,location);
		}
		// get starting contour positions
		HashMap originalContourPositions = new HashMap();
		Iterator contourIt1 = contours.iterator();
		while (contourIt1.hasNext()) {
			ConcreteContour cp = (ConcreteContour)contourIt1.next();
			Polygon p = cp.getPolygon();
			Vector originalPoints = FastHillClimber.getPolygonPoints(p);
			originalContourPositions.put(p,originalPoints);
		}

		// try all possible phases
		int phaseStage = 0;
		boolean fail = false;
		while(phaseStage < deltaX.length && !fail) {

			// move feet according to phases
			Iterator feetIt2 = feet.iterator();
			while (feetIt2.hasNext()) {
				ConcreteSpiderFoot csf = (ConcreteSpiderFoot) feetIt2.next();
				Integer footPhase = (Integer)feetPhases.get(csf);
				int deltaInd = (phaseStage+footPhase.intValue())%deltaX.length;
				csf.setLocation(csf.getX()+deltaX[deltaInd]*circleSize,csf.getY()+deltaY[deltaInd]*circleSize);
			}
			// move contours according to phases
			Iterator contourIt2 = contours.iterator();
			while (contourIt2.hasNext()) {
				ConcreteContour cp = (ConcreteContour) contourIt2.next();
				Polygon p = cp.getPolygon();
				Integer polyPhase = (Integer)contourPhases.get(cp);
				int deltaInd = (phaseStage+polyPhase.intValue())%deltaX.length;
				FastHillClimber.movePolygonPoints(p, deltaX[deltaInd]*circleSize, deltaY[deltaInd]*circleSize);
			}

			// check structure remains consistent after move
			Iterator itUD2 = unitaryDiags.iterator();
			while (itUD2.hasNext()) {
				ConcreteUnitaryDiagram cud = (ConcreteUnitaryDiagram)itUD2.next();
				cud.refreshZones(false);
				cud.notifyObservers();
				List checkers = (List)structureCheckers.get(cud);
				for(int i=0; i<checkers.size(); i++) {
					StructureChecker sc = (StructureChecker)checkers.get(i);
					if(!sc.checkStillOK(cud)) {
						fail = true;
						break;
					};
				}
			}
			
			phaseStage++;
		}

		// reset original feet positions
		Iterator feetIt3 = feet.iterator();
		while (feetIt3.hasNext()) {
			ConcreteSpiderFoot csf = (ConcreteSpiderFoot)feetIt3.next();
			Point originalFootLocation = (Point)originalFeetPositions.get(csf);
			csf.setLocation((int)originalFootLocation.getX(),(int)originalFootLocation.getY());
		}
		// reset original contour positions
		Iterator contourIt3 = contours.iterator();
		while (contourIt3.hasNext()) {
			ConcreteContour cp = (ConcreteContour)contourIt3.next();
			Polygon p = cp.getPolygon();
			Vector originalContourPoints = (Vector)originalContourPositions.get(p);
			FastHillClimber.setPolygonPoints(originalContourPoints,p);
			cp.autoPlaceLabel();
		}

		return !fail;
	}
}
